# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from enum import Enum
import logging
from spinn_front_end_common.utility_models import MultiCastCommand
from spinn_front_end_common.utilities.exceptions import ConfigurationException

logger = logging.getLogger(__name__)

# structure of command is KKKKKKKKKKKKKKKKKKKKK-IIIIIII-F-DDD
# K = ignored key at the top of the command
# I = instruction
# F = format
# D = device
_OFFSET_TO_IGNORED_KEY = 11
_OFFSET_TO_I = 4
_OFFSET_TO_F = 3
_OFFSET_TO_D = 0
_KEY_MASK = 0xFFFFF800
_I_MASK = 0x7F0
_F_MASK = 0x8
_D_MASK = 0x7


# UART masks and shifts
RETINA_UART_SHIFT = 3 + _OFFSET_TO_I
RETINA_WITHOUT_UART_MASK = 0x670
PUSH_BOT_LASER_LED_SPEAKER_FREQUENCY_UART_SHIFT = 1 + _OFFSET_TO_D
PUSH_BOT_MOTOR_WITHOUT_UART_MASK = 0x7C0
PUSH_BOT_MOTOR_UART_SHIFT = 0 + _OFFSET_TO_I


def munich_key(I, F, D):
    return (I << _OFFSET_TO_I) | (F << _OFFSET_TO_F) | (D << _OFFSET_TO_D)


def munich_key_i_d(I, D):
    return munich_key(I, 0, D)


def munich_key_i(I):
    return munich_key(I, 0, 0)


def get_munich_i(key):
    return key & _I_MASK


def get_munich_f(key):
    return key & _F_MASK


def get_munich_d(key):
    return key & _D_MASK


def get_retina_i(key):
    return key & RETINA_WITHOUT_UART_MASK


def get_push_bot_laser_led_speaker_frequency_i(key):
    return get_munich_i(key)


def get_push_bot_motor_i(key):
    return key & PUSH_BOT_MOTOR_WITHOUT_UART_MASK


# Specific fields in the key
_SENSOR_OUTGOING_OFFSET_TO_D = 2
_SENSOR_OUTGOING_OFFSET_TO_I = 7

# Payload fields
_PAYLOAD_RETINA_PAYLOAD_OFFSET = 29
_PAYLOAD_RETINA_PAYLOAD_MASK = 0xE0000000
_PAYLOAD_RETINA_KEY_OFFSET = 26
_PAYLOAD_RETINA_KEY_MASK = 0x1C000000
_PAYLOAD_SENSOR_ID_OFFSET = 27
_PAYLOAD_OFFSET_FOR_SENSOR_TIME = 0


def GET_RETINA_KEY_VALUE(payload):
    return (payload & _PAYLOAD_RETINA_KEY_MASK) >> _PAYLOAD_RETINA_KEY_OFFSET


def GET_RETINA_PAYLOAD_VALUE(payload):
    return (
        (payload & _PAYLOAD_RETINA_PAYLOAD_MASK) >>
        _PAYLOAD_RETINA_PAYLOAD_OFFSET
    )


# command key for setting up the master key of the board
CONFIGURE_MASTER_KEY = munich_key_i_d(127, 0)

# command key for setting up what mode of device running on the board
CHANGE_MODE = munich_key_i_d(127, 1)

# command for turning off retina output
DISABLE_RETINA_EVENT_STREAMING = munich_key_i_d(0, 0)

# command for retina where payload is events
ACTIVE_RETINA_EVENT_STREAMING_KEYS_CONFIGURATION = munich_key_i_d(0, 1)

# command for retina where events are the key
ACTIVE_RETINA_EVENT_STREAMING_SET_KEY = munich_key_i_d(0, 2)

# set timer / counter for timestamps
SET_TIMER_COUNTER_FOR_TIMESTAMPS = munich_key_i_d(0, 3)

# handle master / slave time sync
MASTER_SLAVE_KEY = munich_key_i_d(0, 4)

# command for setting bias (whatever the check that is)
BIAS_KEY = munich_key_i_d(0, 5)

# reset retina key.
RESET_RETINA_KEY = munich_key_i_d(0, 7)

# request on-board sensor data
SENSOR_REPORTING_OFF_KEY = munich_key_i_d(1, 0)

# poll sensors once
POLL_SENSORS_ONCE_KEY = munich_key_i_d(1, 1)

# poll sensors continuously
POLL_SENSORS_CONTINUOUSLY_KEY = munich_key_i_d(1, 2)

# disable motor
ENABLE_DISABLE_MOTOR_KEY = munich_key_i_d(2, 0)

# run motor for total period
MOTOR_RUN_FOR_PERIOD_KEY = munich_key_i_d(2, 1)

# raw output for motor 0 (permanent)
MOTOR_0_RAW_PERM_KEY = munich_key_i_d(2, 4)

# raw output for motor 1 (permanent)
MOTOR_1_RAW_PERM_KEY = munich_key_i_d(2, 5)

# raw output for motor 0 (leak towards 0)
MOTOR_0_RAW_LEAK_KEY = munich_key_i_d(2, 6)

# raw output for motor 1 (leak towards 0)
MOTOR_1_RAW_LEAK_KEY = munich_key_i_d(2, 7)

# motor output duration timer period
MOTOR_TIMER_A_TOTAL_PERIOD_KEY = munich_key_i_d(3, 0)
MOTOR_TIMER_B_TOTAL_PERIOD_KEY = munich_key_i_d(3, 2)
MOTOR_TIMER_C_TOTAL_PERIOD_KEY = munich_key_i_d(3, 4)

# motor output ratio active period
MOTOR_TIMER_A_CHANNEL_0_ACTIVE_PERIOD_KEY = munich_key_i_d(4, 0)
MOTOR_TIMER_A_CHANNEL_1_ACTIVE_PERIOD_KEY = munich_key_i_d(4, 1)
MOTOR_TIMER_B_CHANNEL_0_ACTIVE_PERIOD_KEY = munich_key_i_d(4, 2)
MOTOR_TIMER_B_CHANNEL_1_ACTIVE_PERIOD_KEY = munich_key_i_d(4, 3)
MOTOR_TIMER_C_CHANNEL_0_ACTIVE_PERIOD_KEY = munich_key_i_d(4, 4)
MOTOR_TIMER_C_CHANNEL_1_ACTIVE_PERIOD_KEY = munich_key_i_d(4, 5)

# digital IO Signals
QUERY_STATES_LINES_KEY = munich_key_i_d(5, 0)

# set output pattern to payload
SET_OUTPUT_PATTERN_KEY = munich_key_i_d(5, 1)

# add payload (logic or (PL)) to current output
ADD_PAYLOAD_TO_CURRENT_OUTPUT_KEY = munich_key_i_d(5, 2)

# remove payload (logic or (PL)) to current output from current output
REMOVE_PAYLOAD_TO_CURRENT_OUTPUT_KEY = munich_key_i_d(5, 3)

# set payload pins to high impedance
SET_PAYLOAD_TO_HIGH_IMPEDANCE_KEY = munich_key_i_d(5, 4)

# set laser params for PushBot
PUSH_BOT_LASER_CONFIG_TOTAL_PERIOD = munich_key_i_d(3, 0)
PUSH_BOT_LASER_CONFIG_ACTIVE_TIME = munich_key_i_d(4, 0)
PUSH_BOT_LASER_FREQUENCY = munich_key_i_d(37, 1)

# set led params for PushBot
PUSH_BOT_LED_CONFIG_TOTAL_PERIOD = munich_key_i_d(3, 4)
PUSH_BOT_LED_BACK_CONFIG_ACTIVE_TIME = munich_key_i_d(4, 4)
PUSH_BOT_LED_FRONT_CONFIG_ACTIVE_TIME = munich_key_i_d(4, 5)
PUSH_BOT_LED_FREQUENCY = munich_key_i_d(37, 0)

# set speaker params for PushBot
PUSH_BOT_SPEAKER_CONFIG_TOTAL_PERIOD = munich_key_i_d(3, 2)
PUSH_BOT_SPEAKER_CONFIG_ACTIVE_TIME = munich_key_i_d(4, 2)
PUSH_BOT_SPEAKER_TONE_BEEP = munich_key_i_d(36, 0)
PUSH_BOT_SPEAKER_TONE_MELODY = munich_key_i_d(36, 1)

# PushBot motor control
PUSH_BOT_MOTOR_0_PERMANENT_VELOCITY = munich_key_i_d(32, 0)
PUSH_BOT_MOTOR_1_PERMANENT_VELOCITY = munich_key_i_d(32, 1)
PUSH_BOT_MOTOR_0_LEAKY_VELOCITY = munich_key_i_d(32, 2)
PUSH_BOT_MOTOR_1_LEAKY_VELOCITY = munich_key_i_d(32, 3)

# payload for master slave
_PAYLOAD_MASTER_SLAVE_USE_INTERNAL_COUNTER = 0
_PAYLOAD_MASTER_SLAVE_SET_SLAVE = 1
_PAYLOAD_MASTER_SLAVE_SET_MASTER_CLOCK_NOT_STARTED = 2
_PAYLOAD_MASTER_SLAVE_SET_MASTER_CLOCK_ACTIVE = 4


class RetinaKey(Enum):
    FIXED_KEY = (0, 128, 7)
    NATIVE_128_X_128 = (1, 128, 7)
    DOWNSAMPLE_64_X_64 = (2, 64, 6)
    DOWNSAMPLE_32_X_32 = (3, 32, 5)
    DOWNSAMPLE_16_X_16 = (4, 16, 4)

    def __init__(self, value, pixels, bits_per_coordinate):
        self._value_ = value << _PAYLOAD_RETINA_KEY_OFFSET
        self._pixels = pixels
        self._bits_per_coordinate = bits_per_coordinate

    @property
    def n_neurons(self):
        return 2 * (self._pixels ** 2)

    @property
    def pixels(self):
        return self._pixels

    @property
    def bits_per_coordinate(self):
        return self._bits_per_coordinate


class RetinaPayload(Enum):
    NO_PAYLOAD = (0, 0)
    EVENTS_IN_PAYLOAD = (0, 4)
    DELTA_TIMESTAMPS = (1, 4)
    ABSOLUTE_2_BYTE_TIMESTAMPS = (2, 2)
    ABSOLUTE_3_BYTE_TIMESTAMPS = (3, 3)
    ABSOLUTE_4_BYTE_TIMESTAMPS = (4, 4)

    def __init__(self, value, n_payload_bytes):
        self._value_ = value << _PAYLOAD_RETINA_PAYLOAD_OFFSET
        self._n_payload_bytes = n_payload_bytes

    @property
    def n_payload_bytes(self):
        return self._n_payload_bytes


class MunichIoSpiNNakerLinkProtocol(object):
    """ Provides Multicast commands for the Munich SpiNNaker-Link protocol
    """
    __slots__ = ["__instance_key", "__mode", "__uart_id"]

    # types of modes supported by this protocol
    MODES = Enum(
        value="MODES",
        names=[('RESET_TO_DEFAULT', 0),
               ('PUSH_BOT', 1),
               ('SPOMNIBOT', 2),
               ('BALL_BALANCER', 3),
               ('MY_ORO_BOTICS', 4),
               ('FREE', 5)])

    # The instance of the protocol in use, to ensure that each vertex that is
    # to send commands to the PushBot uses a different outgoing key; the top
    # part of the key is ignored, so this works out!
    protocol_instance = 0

    # Keeps track of whether the mode has been configured already
    __sent_mode_command = False

    def __init__(self, mode, instance_key=None, uart_id=0):
        """

        :param mode: The mode of operation of the protocol
        :param instance_key: The optional instance key to use
        :param uart_id: The ID of the UART when needed
        """
        self.__mode = mode

        # Create a key for this instance of the protocol
        # - see above for reasoning
        if instance_key is None:
            self.__instance_key = (
                MunichIoSpiNNakerLinkProtocol.protocol_instance <<
                _OFFSET_TO_IGNORED_KEY
            )
            MunichIoSpiNNakerLinkProtocol.protocol_instance += 1
        else:
            self.__instance_key = instance_key

        self.__uart_id = uart_id

    @property
    def mode(self):
        return self.__mode

    @property
    def uart_id(self):
        return self.__uart_id

    @staticmethod
    def sent_mode_command():
        """ True if the mode command has ever been requested by any instance
        """
        return MunichIoSpiNNakerLinkProtocol.__sent_mode_command

    @property
    def instance_key(self):
        """ The key of this instance of the protocol
        """
        return self.__instance_key

    def _get_key(self, command, offset_to_uart_id=None):
        if offset_to_uart_id is None:
            return command | self.__instance_key
        return (
            command | self.__instance_key |
            (self.__uart_id << offset_to_uart_id)
        )

    @property
    def configure_master_key_key(self):
        return self._get_key(CONFIGURE_MASTER_KEY)

    def configure_master_key(self, new_key, time=None):
        return MultiCastCommand(
            key=self.configure_master_key_key, payload=new_key, time=time)

    @property
    def set_mode_key(self):
        return self._get_key(CHANGE_MODE)

    def set_mode(self, time=None):
        MunichIoSpiNNakerLinkProtocol.__sent_mode_command = True
        return MultiCastCommand(
            key=self.set_mode_key, payload=self.__mode.value, time=time)

    @property
    def set_retina_key_key(self):
        return self._get_key(
            ACTIVE_RETINA_EVENT_STREAMING_SET_KEY, RETINA_UART_SHIFT)

    def set_retina_key(self, new_key, time=None):
        return MultiCastCommand(
            key=self.set_retina_key_key,
            payload=new_key, time=time)

    @property
    def disable_retina_key(self):
        return self._get_key(DISABLE_RETINA_EVENT_STREAMING, RETINA_UART_SHIFT)

    def disable_retina(self, time=None):
        return MultiCastCommand(key=self.disable_retina_key, time=time)

    @property
    def master_slave_key(self):
        return self._get_key(MASTER_SLAVE_KEY, RETINA_UART_SHIFT)

    def master_slave_use_internal_counter(self, time=None):
        return MultiCastCommand(
            key=self.master_slave_key,
            payload=_PAYLOAD_MASTER_SLAVE_USE_INTERNAL_COUNTER, time=time)

    def master_slave_set_slave(self, time=None):
        return MultiCastCommand(
            key=self.master_slave_key,
            payload=_PAYLOAD_MASTER_SLAVE_SET_SLAVE, time=time)

    def master_slave_set_master_clock_not_started(self, time=None):
        return MultiCastCommand(
            key=self.master_slave_key,
            payload=_PAYLOAD_MASTER_SLAVE_SET_MASTER_CLOCK_NOT_STARTED,
            time=time)

    def master_slave_set_master_clock_active(self, time=None):
        return MultiCastCommand(
            key=self.master_slave_key,
            payload=_PAYLOAD_MASTER_SLAVE_SET_MASTER_CLOCK_ACTIVE,
            time=time)

    @property
    def bias_values_key(self):
        return self._get_key(BIAS_KEY, RETINA_UART_SHIFT)

    def bias_values(self, bias_id, bias_value, time=None):
        return MultiCastCommand(
            key=self.bias_values_key,
            payload=((bias_id << 0) | (bias_value << 8)), time=time)

    @property
    def reset_retina_key(self):
        return self._get_key(RESET_RETINA_KEY, RETINA_UART_SHIFT)

    def reset_retina(self, time=None):
        return MultiCastCommand(
            key=self.reset_retina_key, time=time)

    @property
    def turn_off_sensor_reporting_key(self):
        return self._get_key(SENSOR_REPORTING_OFF_KEY)

    def turn_off_sensor_reporting(self, sensor_id, time=None):
        return MultiCastCommand(
            key=self.turn_off_sensor_reporting_key,
            payload=(sensor_id << _PAYLOAD_SENSOR_ID_OFFSET), time=time)

    @property
    def poll_sensors_once_key(self):
        return self._get_key(POLL_SENSORS_ONCE_KEY)

    def poll_sensors_once(self, sensor_id, time=None):
        return MultiCastCommand(
            key=self.poll_sensors_once_key,
            payload=(sensor_id << _PAYLOAD_SENSOR_ID_OFFSET), time=time)

    @property
    def poll_individual_sensor_continuously_key(self):
        return self._get_key(POLL_SENSORS_CONTINUOUSLY_KEY)

    def poll_individual_sensor_continuously(
            self, sensor_id, time_in_ms, time=None):
        return MultiCastCommand(
            key=self.poll_individual_sensor_continuously_key,
            payload=((sensor_id << _PAYLOAD_SENSOR_ID_OFFSET) |
                     (time_in_ms << _PAYLOAD_OFFSET_FOR_SENSOR_TIME)),
            time=time)

    @property
    def enable_disable_motor_key(self):
        return self._get_key(ENABLE_DISABLE_MOTOR_KEY, RETINA_UART_SHIFT)

    def generic_motor_enable(self, time=None):
        return MultiCastCommand(
            key=self.enable_disable_motor_key, payload=1, time=time)

    def generic_motor_disable(self, time=None):
        return MultiCastCommand(
            key=self.enable_disable_motor_key, payload=0, time=time)

    @property
    def generic_motor_total_period_key(self):
        return self._get_key(MOTOR_RUN_FOR_PERIOD_KEY, RETINA_UART_SHIFT)

    def generic_motor_total_period(
            self, time_in_ms, uart_id=0, time=None):
        return MultiCastCommand(
            key=self.generic_motor_total_period_key,
            payload=time_in_ms, time=time)

    @property
    def generic_motor0_raw_output_permanent_key(self):
        return self._get_key(MOTOR_0_RAW_PERM_KEY, RETINA_UART_SHIFT)

    def generic_motor0_raw_output_permanent(self, pwm_signal, time=None):
        return MultiCastCommand(
            key=self.generic_motor0_raw_output_permanent_key,
            payload=pwm_signal, time=time)

    @property
    def generic_motor1_raw_output_permanent_key(self):
        return self._get_key(MOTOR_1_RAW_PERM_KEY, RETINA_UART_SHIFT)

    def generic_motor1_raw_output_permanent(self, pwm_signal, time=None):
        return MultiCastCommand(
            key=self.generic_motor1_raw_output_permanent_key,
            payload=pwm_signal, time=time)

    @property
    def generic_motor0_raw_output_leak_to_0_key(self):
        return self._get_key(MOTOR_0_RAW_LEAK_KEY, RETINA_UART_SHIFT)

    def generic_motor0_raw_output_leak_to_0(self, pwm_signal, time=None):
        return MultiCastCommand(
            key=self.generic_motor0_raw_output_leak_to_0_key,
            payload=pwm_signal, time=time)

    @property
    def generic_motor1_raw_output_leak_to_0_key(self):
        return self._get_key(MOTOR_1_RAW_LEAK_KEY, RETINA_UART_SHIFT)

    def generic_motor1_raw_output_leak_to_0(self, pwm_signal, time=None):
        return MultiCastCommand(
            key=self.generic_motor1_raw_output_leak_to_0_key,
            payload=pwm_signal, time=time)

    @property
    def pwm_pin_output_timer_a_duration_key(self):
        return self._get_key(MOTOR_TIMER_A_TOTAL_PERIOD_KEY, RETINA_UART_SHIFT)

    def pwm_pin_output_timer_a_duration(self, timer_period, time=None):
        return MultiCastCommand(
            key=self.pwm_pin_output_timer_a_duration_key,
            payload=timer_period, time=time)

    @property
    def pwm_pin_output_timer_b_duration_key(self):
        return self._get_key(MOTOR_TIMER_B_TOTAL_PERIOD_KEY, RETINA_UART_SHIFT)

    def pwm_pin_output_timer_b_duration(self, timer_period, time=None):
        return MultiCastCommand(
            key=self.pwm_pin_output_timer_b_duration_key,
            payload=timer_period, time=time)

    @property
    def pwm_pin_output_timer_c_duration_key(self):
        return self._get_key(MOTOR_TIMER_C_TOTAL_PERIOD_KEY, RETINA_UART_SHIFT)

    def pwm_pin_output_timer_c_duration(self, timer_period, time=None):
        return MultiCastCommand(
            key=self.pwm_pin_output_timer_c_duration_key,
            payload=timer_period, time=time)

    @property
    def pwm_pin_output_timer_a_channel_0_ratio_key(self):
        return self._get_key(
            MOTOR_TIMER_A_CHANNEL_0_ACTIVE_PERIOD_KEY, RETINA_UART_SHIFT)

    def pwm_pin_output_timer_a_channel_0_ratio(self, timer_period, time=None):
        return MultiCastCommand(
            key=self.pwm_pin_output_timer_a_channel_0_ratio_key,
            payload=timer_period, time=time)

    @property
    def pwm_pin_output_timer_a_channel_1_ratio_key(self):
        return self._get_key(
            MOTOR_TIMER_A_CHANNEL_1_ACTIVE_PERIOD_KEY, RETINA_UART_SHIFT)

    def pwm_pin_output_timer_a_channel_1_ratio(self, timer_period, time=None):
        return MultiCastCommand(
            key=self.pwm_pin_output_timer_a_channel_1_ratio_key,
            payload=timer_period, time=time)

    def pwm_pin_output_timer_b_channel_0_ratio_key(self):
        return self._get_key(
            MOTOR_TIMER_B_CHANNEL_0_ACTIVE_PERIOD_KEY, RETINA_UART_SHIFT)

    def pwm_pin_output_timer_b_channel_0_ratio(self, timer_period, time=None):
        return MultiCastCommand(
            key=self.pwm_pin_output_timer_b_channel_0_ratio_key,
            payload=timer_period, time=time)

    @property
    def pwm_pin_output_timer_b_channel_1_ratio_key(self):
        return self._get_key(
            MOTOR_TIMER_B_CHANNEL_1_ACTIVE_PERIOD_KEY, RETINA_UART_SHIFT)

    def pwm_pin_output_timer_b_channel_1_ratio(self, timer_period, time=None):
        return MultiCastCommand(
            key=self.pwm_pin_output_timer_b_channel_1_ratio_key,
            payload=timer_period, time=time)

    @property
    def pwm_pin_output_timer_c_channel_0_ratio_key(self):
        return self._get_key(
            MOTOR_TIMER_C_CHANNEL_0_ACTIVE_PERIOD_KEY, RETINA_UART_SHIFT)

    def pwm_pin_output_timer_c_channel_0_ratio(self, timer_period, time=None):
        return MultiCastCommand(
            key=self.pwm_pin_output_timer_c_channel_0_ratio_key,
            payload=timer_period, time=time)

    def pwm_pin_output_timer_c_channel_1_ratio_key(self):
        return self._get_key(
            MOTOR_TIMER_C_CHANNEL_1_ACTIVE_PERIOD_KEY, RETINA_UART_SHIFT)

    def pwm_pin_output_timer_c_channel_1_ratio(self, timer_period, time=None):
        return MultiCastCommand(
            key=self.pwm_pin_output_timer_c_channel_1_ratio_key,
            payload=timer_period, time=time)

    @property
    def query_state_of_io_lines_key(self):
        return self._get_key(QUERY_STATES_LINES_KEY)

    def query_state_of_io_lines(self, time=None):
        return MultiCastCommand(
            key=self.query_state_of_io_lines_key, time=time)

    @property
    def set_output_pattern_for_payload_key(self):
        return self._get_key(SET_OUTPUT_PATTERN_KEY)

    def set_output_pattern_for_payload(self, payload, time=None):
        return MultiCastCommand(
            key=self.set_output_pattern_for_payload_key, payload=payload,
            time=time)

    @property
    def add_payload_logic_to_current_output_key(self):
        return self._get_key(ADD_PAYLOAD_TO_CURRENT_OUTPUT_KEY)

    def add_payload_logic_to_current_output(self, payload, time=None):
        return MultiCastCommand(
            key=self.add_payload_logic_to_current_output_key,
            payload=payload, time=time)

    @property
    def remove_payload_logic_to_current_output_key(self):
        return self._get_key(REMOVE_PAYLOAD_TO_CURRENT_OUTPUT_KEY)

    def remove_payload_logic_to_current_output(self, payload, time=None):
        return MultiCastCommand(
            key=self.remove_payload_logic_to_current_output_key,
            payload=payload, time=time)

    @property
    def set_payload_pins_to_high_impedance_key(self):
        return self._get_key(SET_PAYLOAD_TO_HIGH_IMPEDANCE_KEY)

    def set_payload_pins_to_high_impedance(self, payload, time=None):
        return MultiCastCommand(
            key=self.set_payload_pins_to_high_impedance_key,
            payload=payload, time=time)

    def _check_for_pushbot_mode(self):
        if self.__mode is not self.MODES.PUSH_BOT:
            raise ConfigurationException(
                "The mode you configured is not the PushBot, and so this "
                "message is invalid for mode {}".format(self.__mode))

    @property
    def push_bot_laser_config_total_period_key(self):
        return self._get_key(
            PUSH_BOT_LASER_CONFIG_TOTAL_PERIOD, RETINA_UART_SHIFT)

    def push_bot_laser_config_total_period(self, total_period, time=None):
        self._check_for_pushbot_mode()
        return MultiCastCommand(
            key=self.push_bot_laser_config_total_period_key,
            payload=total_period, time=time)

    @property
    def push_bot_laser_config_active_time_key(self):
        return self._get_key(
            PUSH_BOT_LASER_CONFIG_ACTIVE_TIME, RETINA_UART_SHIFT)

    def push_bot_laser_config_active_time(self, active_time, time=None):
        self._check_for_pushbot_mode()
        return MultiCastCommand(
            key=self.push_bot_laser_config_active_time_key,
            payload=active_time, time=time)

    @property
    def push_bot_laser_set_frequency_key(self):
        return self._get_key(
            PUSH_BOT_LASER_FREQUENCY,
            PUSH_BOT_LASER_LED_SPEAKER_FREQUENCY_UART_SHIFT)

    def push_bot_laser_set_frequency(self, frequency, time=None):
        self._check_for_pushbot_mode()
        return MultiCastCommand(
            key=self.push_bot_laser_set_frequency_key,
            payload=frequency, time=time)

    @property
    def push_bot_speaker_config_total_period_key(self):
        return self._get_key(
            PUSH_BOT_SPEAKER_CONFIG_TOTAL_PERIOD, RETINA_UART_SHIFT)

    def push_bot_speaker_config_total_period(
            self, total_period, time=None):
        self._check_for_pushbot_mode()
        return MultiCastCommand(
            key=self.push_bot_speaker_config_total_period_key,
            payload=total_period, time=time)

    @property
    def push_bot_speaker_config_active_time_key(self):
        return self._get_key(
            PUSH_BOT_SPEAKER_CONFIG_ACTIVE_TIME, RETINA_UART_SHIFT)

    def push_bot_speaker_config_active_time(self, active_time, time=None):
        self._check_for_pushbot_mode()
        return MultiCastCommand(
            key=self.push_bot_speaker_config_active_time_key,
            payload=active_time, time=time)

    @property
    def push_bot_speaker_set_tone_key(self):
        return self._get_key(
            PUSH_BOT_SPEAKER_TONE_BEEP,
            PUSH_BOT_LASER_LED_SPEAKER_FREQUENCY_UART_SHIFT)

    def push_bot_speaker_set_tone(self, frequency, time=None):
        self._check_for_pushbot_mode()
        return MultiCastCommand(
            key=self.push_bot_speaker_set_tone_key,
            payload=frequency, time=time)

    @property
    def push_bot_speaker_set_melody_key(self):
        return self._get_key(
            PUSH_BOT_SPEAKER_TONE_MELODY,
            PUSH_BOT_LASER_LED_SPEAKER_FREQUENCY_UART_SHIFT)

    def push_bot_speaker_set_melody(self, melody, time=None):
        self._check_for_pushbot_mode()
        return MultiCastCommand(
            key=self.push_bot_speaker_set_melody_key,
            payload=melody, time=time)

    @property
    def push_bot_led_total_period_key(self):
        return self._get_key(
            PUSH_BOT_LED_CONFIG_TOTAL_PERIOD, RETINA_UART_SHIFT)

    def push_bot_led_total_period(self, total_period, time=None):
        self._check_for_pushbot_mode()
        return MultiCastCommand(
            key=self.push_bot_led_total_period_key,
            payload=total_period, time=time)

    @property
    def push_bot_led_back_active_time_key(self):
        return self._get_key(
            PUSH_BOT_LED_BACK_CONFIG_ACTIVE_TIME, RETINA_UART_SHIFT)

    def push_bot_led_back_active_time(self, active_time, time=None):
        self._check_for_pushbot_mode()
        return MultiCastCommand(
            key=self.push_bot_led_back_active_time_key,
            payload=active_time, time=time)

    @property
    def push_bot_led_front_active_time_key(self):
        return self._get_key(
            PUSH_BOT_LED_FRONT_CONFIG_ACTIVE_TIME, RETINA_UART_SHIFT)

    def push_bot_led_front_active_time(self, active_time, time=None):
        self._check_for_pushbot_mode()
        return MultiCastCommand(
            key=self.push_bot_led_front_active_time_key,
            payload=active_time, time=time)

    @property
    def push_bot_led_set_frequency_key(self):
        return self._get_key(
            PUSH_BOT_LED_FREQUENCY,
            PUSH_BOT_LASER_LED_SPEAKER_FREQUENCY_UART_SHIFT)

    def push_bot_led_set_frequency(self, frequency, time=None):
        self._check_for_pushbot_mode()
        return MultiCastCommand(
            key=self.push_bot_led_set_frequency_key,
            payload=frequency, time=time)

    @property
    def push_bot_motor_0_permanent_key(self):
        return self._get_key(
            PUSH_BOT_MOTOR_0_PERMANENT_VELOCITY, PUSH_BOT_MOTOR_UART_SHIFT)

    def push_bot_motor_0_permanent(self, velocity, time=None):
        self._check_for_pushbot_mode()
        return MultiCastCommand(
            key=self.push_bot_motor_0_permanent_key,
            payload=velocity, time=time)

    @property
    def push_bot_motor_1_permanent_key(self):
        return self._get_key(
            PUSH_BOT_MOTOR_1_PERMANENT_VELOCITY, PUSH_BOT_MOTOR_UART_SHIFT)

    def push_bot_motor_1_permanent(self, velocity, time=None):
        self._check_for_pushbot_mode()
        return MultiCastCommand(
            key=self.push_bot_motor_1_permanent_key,
            payload=velocity, time=time)

    @property
    def push_bot_motor_0_leaking_towards_zero_key(self):
        return self._get_key(
            PUSH_BOT_MOTOR_0_LEAKY_VELOCITY, PUSH_BOT_MOTOR_UART_SHIFT)

    def push_bot_motor_0_leaking_towards_zero(self, velocity, time=None):
        self._check_for_pushbot_mode()
        return MultiCastCommand(
            key=self.push_bot_motor_0_leaking_towards_zero_key,
            payload=velocity, time=time)

    @property
    def push_bot_motor_1_leaking_towards_zero_key(self):
        return self._get_key(
            PUSH_BOT_MOTOR_1_LEAKY_VELOCITY, PUSH_BOT_MOTOR_UART_SHIFT)

    def push_bot_motor_1_leaking_towards_zero(self, velocity, time=None):
        self._check_for_pushbot_mode()
        return MultiCastCommand(
            key=self.push_bot_motor_1_leaking_towards_zero_key,
            payload=velocity, time=time)

    def sensor_transmission_key(self, sensor_id):
        return ((sensor_id << _SENSOR_OUTGOING_OFFSET_TO_D) |
                (self.__uart_id << _SENSOR_OUTGOING_OFFSET_TO_I))

    @property
    def set_retina_transmission_key(self):
        return self._get_key(
            ACTIVE_RETINA_EVENT_STREAMING_KEYS_CONFIGURATION,
            RETINA_UART_SHIFT)

    def set_retina_transmission(
            self, retina_key=RetinaKey.NATIVE_128_X_128,
            retina_payload=None, time=None):
        """ Set the retina transmission key

        :param retina_key: the new key for the retina
        :param retina_payload: \
            the new payload for the set retina key command packet
        :type retina_payload: enum or None
        :param time: when to transmit this packet
        :return: the command to send
        :rtype: \
            :py:class:`spinn_front_end_common.utility_models.multi_cast_command.MultiCastCommand`
        """

        if retina_key == RetinaKey.FIXED_KEY and retina_payload is None:
            retina_payload = RetinaPayload.EVENTS_IN_PAYLOAD

        if retina_payload is None:
            retina_payload = RetinaPayload.NO_PAYLOAD

        if (retina_key == RetinaKey.FIXED_KEY and
                retina_payload != RetinaPayload.EVENTS_IN_PAYLOAD):
            raise ConfigurationException(
                "If the Retina Key is FIXED_KEY, the payload must be"
                " EVENTS_IN_PAYLOAD")

        return MultiCastCommand(
            key=self.set_retina_transmission_key,
            payload=retina_key.value | retina_payload.value,
            time=time)



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from .abstract_neuron_recordable import AbstractNeuronRecordable
from .abstract_spike_recordable import AbstractSpikeRecordable
from .eieio_spike_recorder import EIEIOSpikeRecorder
from .neuron_recorder import NeuronRecorder
from .multi_spike_recorder import MultiSpikeRecorder
from .recording_utils import (
    get_buffer_sizes, get_data, get_recording_region_size_in_bytes,
    needs_buffering, pull_off_cached_lists)
from .simple_population_settable import SimplePopulationSettable

__all__ = ["AbstractNeuronRecordable", "AbstractSpikeRecordable",
           "EIEIOSpikeRecorder", "NeuronRecorder", "MultiSpikeRecorder",
           "SimplePopulationSettable", "get_buffer_sizes", "get_data",
           "needs_buffering", "get_recording_region_size_in_bytes",
           "pull_off_cached_lists", ]



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from spinn_utilities.overrides import overrides
from spinn_front_end_common.abstract_models import (
    AbstractSendMeMulticastCommandsVertex)
from spinn_front_end_common.abstract_models.impl import (
    ProvidesKeyToAtomMappingImpl)
from spinn_front_end_common.utilities.exceptions import ConfigurationException
from .push_bot_ethernet_device import PushBotEthernetDevice
from spynnaker.pyNN.external_devices_models.push_bot.push_bot_parameters \
    import (
        PushBotLaser)


class PushBotEthernetLaserDevice(
        PushBotEthernetDevice, AbstractSendMeMulticastCommandsVertex,
        ProvidesKeyToAtomMappingImpl):
    """ The Laser of a PushBot
    """

    def __init__(
            self, laser, protocol,
            start_active_time=None, start_total_period=None,
            start_frequency=None, timesteps_between_send=None):
        """

        :param laser: The PushBotLaser value to control
        :param protocol: The protocol instance to get commands from
        :param start_active_time: The "active time" value to send at the start
        :param start_total_period:\
            The "total period" value to send at the start
        :param start_frequency: The "frequency" to send at the start
        :param timesteps_between_send:\
            The number of timesteps between sending commands to the device,\
            or None to use the default
        """
        # pylint: disable=too-many-arguments
        if not isinstance(laser, PushBotLaser):
            raise ConfigurationException(
                "laser parameter must be a PushBotLaser value")

        super(PushBotEthernetLaserDevice, self).__init__(
            protocol, laser, True, timesteps_between_send)

        # protocol specific data items
        self.__command_protocol = protocol
        self.__start_active_time = start_active_time
        self.__start_total_period = start_total_period
        self.__start_frequency = start_frequency

    def set_command_protocol(self, command_protocol):
        self.__command_protocol = command_protocol

    @property
    @overrides(AbstractSendMeMulticastCommandsVertex.start_resume_commands)
    def start_resume_commands(self):
        commands = list()

        # add mode command if not done already
        if not self.protocol.sent_mode_command():
            commands.append(self.protocol.set_mode())

        # device specific commands
        if self.__start_total_period is not None:
            commands.append(
                self.__command_protocol.push_bot_laser_config_total_period(
                    total_period=self.__start_total_period))
        if self.__start_active_time is not None:
            commands.append(
                self.__command_protocol.push_bot_laser_config_active_time(
                    active_time=self.__start_active_time))
        if self.__start_frequency is not None:
            commands.append(
                self.__command_protocol.push_bot_laser_set_frequency(
                    frequency=self.__start_frequency))
        return commands

    @property
    @overrides(AbstractSendMeMulticastCommandsVertex.pause_stop_commands)
    def pause_stop_commands(self):
        return [
            self.__command_protocol.push_bot_laser_config_total_period(0),
            self.__command_protocol.push_bot_laser_config_active_time(0),
            self.__command_protocol.push_bot_laser_set_frequency(0)]

    @property
    @overrides(AbstractSendMeMulticastCommandsVertex.timed_commands)
    def timed_commands(self):
        return []



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from pacman.model.graphs.application import ApplicationSpiNNakerLinkVertex
from spinn_front_end_common.abstract_models.impl import (
    ProvidesKeyToAtomMappingImpl)


class ExternalCochleaDevice(
        ApplicationSpiNNakerLinkVertex, ProvidesKeyToAtomMappingImpl):
    __slots__ = []

    def __init__(
            self, n_neurons, spinnaker_link, label=None, board_address=None):
        super(ExternalCochleaDevice, self).__init__(
            n_atoms=n_neurons, spinnaker_link_id=spinnaker_link,
            label=label, max_atoms_per_core=n_neurons,
            board_address=board_address)



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from spinn_utilities.overrides import overrides
from pacman.model.graphs.application import ApplicationEdge
from .delayed_machine_edge import DelayedMachineEdge


class DelayedApplicationEdge(ApplicationEdge):
    __slots__ = [
        "__synapse_information"]

    def __init__(
            self, pre_vertex, post_vertex, synapse_information, label=None):
        super(DelayedApplicationEdge, self).__init__(
            pre_vertex, post_vertex, label=label)
        self.__synapse_information = [synapse_information]

    @property
    def synapse_information(self):
        return self.__synapse_information

    def add_synapse_information(self, synapse_information):
        self.__synapse_information.append(synapse_information)

    @overrides(ApplicationEdge.create_machine_edge)
    def create_machine_edge(self, pre_vertex, post_vertex, label):
        return DelayedMachineEdge(
            self.__synapse_information, pre_vertex, post_vertex, label)



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
import math
import numpy
from numpy import (
    arccos, arcsin, arctan, arctan2, ceil, cos, cosh, exp, fabs, floor, fmod,
    hypot, ldexp, log, log10, modf, power, sin, sinh, sqrt, tan, tanh, maximum,
    minimum, e, pi)
from spinn_utilities.overrides import overrides
from spinn_utilities.safe_eval import SafeEval
from spynnaker.pyNN.utilities import utility_calls
from .abstract_connector import AbstractConnector

logger = logging.getLogger(__name__)
# support for arbitrary expression for the distance dependence
_d_expr_context = SafeEval(math, numpy, arccos, arcsin, arctan, arctan2, ceil,
                           cos, cosh, exp, fabs, floor, fmod, hypot, ldexp,
                           log, log10, modf, power, sin, sinh, sqrt, tan, tanh,
                           maximum, minimum, e=e, pi=pi)


class DistanceDependentProbabilityConnector(AbstractConnector):
    """ Make connections using a distribution which varies with distance.
    """

    __slots__ = [
        "__allow_self_connections",
        "__d_expression",
        "__probs"]

    def __init__(
            self, d_expression, allow_self_connections=True, safe=True,
            verbose=False, n_connections=None, rng=None):
        """
        :param d_expression:\
            the right-hand side of a valid python expression for\
            probability, involving 'd', e.g. "exp(-abs(d))", or "d<3",\
            that can be parsed by eval(), that computes the distance\
            dependent distribution.
        :type d_expression: str
        :param allow_self_connections:\
            if the connector is used to connect a Population to itself, this\
            flag determines whether a neuron is allowed to connect to itself,\
            or only to other neurons in the Population.
        :type d_expression: bool
        :param space:\
            a Space object, needed if you wish to specify distance-dependent\
            weights or delays.
        :type space: pyNN.Space
        :param n_connections:\
            The number of efferent synaptic connections per neuron.
        :type n_connections: int or None
        """
        # pylint: disable=too-many-arguments
        super(DistanceDependentProbabilityConnector, self).__init__(
            safe, verbose)
        self.__d_expression = d_expression
        self.__allow_self_connections = allow_self_connections
        self._rng = rng
        if n_connections is not None:
            raise NotImplementedError(
                "n_connections is not implemented for"
                " DistanceDependentProbabilityConnector on this platform")

    @overrides(AbstractConnector.set_projection_information)
    def set_projection_information(
            self, pre_population, post_population, rng, machine_time_step):
        AbstractConnector.set_projection_information(
            self, pre_population, post_population, rng, machine_time_step)
        self._set_probabilities()

    def _set_probabilities(self):
        # Set the probabilities up-front for now
        # TODO: Work out how this can be done statistically
        expand_distances = self._expand_distances(self.__d_expression)
        pre_positions = self.pre_population.positions
        post_positions = self.post_population.positions

        d1 = self.space.distances(
            pre_positions, post_positions, expand_distances)

        # PyNN 0.8 returns a flattened (C-style) array from space.distances,
        # so the easiest thing to do here is to reshape back to the "expected"
        # PyNN 0.7 shape; otherwise later code gets confusing and difficult
        if (len(d1.shape) == 1):
            d = numpy.reshape(d1, (pre_positions.shape[0],
                                   post_positions.shape[0]))
        else:
            d = d1

        self.__probs = _d_expr_context.eval(self.__d_expression, d=d)

    @overrides(AbstractConnector.get_delay_maximum)
    def get_delay_maximum(self, delays):
        return self._get_delay_maximum(
            delays,
            utility_calls.get_probable_maximum_selected(
                self._n_pre_neurons * self._n_post_neurons,
                self._n_pre_neurons * self._n_post_neurons,
                numpy.amax(self.__probs)))

    @overrides(AbstractConnector.get_n_connections_from_pre_vertex_maximum)
    def get_n_connections_from_pre_vertex_maximum(
            self, delays, post_vertex_slice, min_delay=None, max_delay=None):
        # pylint: disable=too-many-arguments
        max_prob = numpy.amax(
            self.__probs[0:self._n_pre_neurons, post_vertex_slice.as_slice])
        n_connections = utility_calls.get_probable_maximum_selected(
            self._n_pre_neurons * self._n_post_neurons,
            post_vertex_slice.n_atoms, max_prob)

        if min_delay is None or max_delay is None:
            return int(math.ceil(n_connections))

        return self._get_n_connections_from_pre_vertex_with_delay_maximum(
            delays, self._n_pre_neurons * self._n_post_neurons,
            n_connections, min_delay, max_delay)

    @overrides(AbstractConnector.get_n_connections_to_post_vertex_maximum)
    def get_n_connections_to_post_vertex_maximum(self):
        # pylint: disable=too-many-arguments
        return utility_calls.get_probable_maximum_selected(
            self._n_pre_neurons * self._n_post_neurons, self._n_post_neurons,
            numpy.amax(self.__probs))

    @overrides(AbstractConnector.get_weight_maximum)
    def get_weight_maximum(self, weights):
        # pylint: disable=too-many-arguments
        return self._get_weight_maximum(
            weights,
            utility_calls.get_probable_maximum_selected(
                self._n_pre_neurons * self._n_post_neurons,
                self._n_pre_neurons * self._n_post_neurons,
                numpy.amax(self.__probs)))

    @overrides(AbstractConnector.create_synaptic_block)
    def create_synaptic_block(
            self, weights, delays, pre_slices, pre_slice_index, post_slices,
            post_slice_index, pre_vertex_slice, post_vertex_slice,
            synapse_type):

        probs = self.__probs[
            pre_vertex_slice.as_slice, post_vertex_slice.as_slice].reshape(-1)
        n_items = pre_vertex_slice.n_atoms * post_vertex_slice.n_atoms
        items = self._rng.next(n_items)

        # If self connections are not allowed, remove the possibility of
        # self connections by setting them to a value of infinity
        if not self.__allow_self_connections:
            items[0:n_items:post_vertex_slice.n_atoms + 1] = numpy.inf

        present = items < probs
        ids = numpy.where(present)[0]
        n_connections = numpy.sum(present)

        block = numpy.zeros(
            n_connections, dtype=self.NUMPY_SYNAPSES_DTYPE)
        block["source"] = (
            (ids // post_vertex_slice.n_atoms) + pre_vertex_slice.lo_atom)
        block["target"] = (
            (ids % post_vertex_slice.n_atoms) + post_vertex_slice.lo_atom)
        block["weight"] = self._generate_weights(
            weights, n_connections, None, pre_vertex_slice, post_vertex_slice)
        block["delay"] = self._generate_delays(
            delays, n_connections, None, pre_vertex_slice, post_vertex_slice)
        block["synapse_type"] = synapse_type
        return block

    def __repr__(self):
        return "DistanceDependentProbabilityConnector({})".format(
            self.__d_expression)

    @property
    def allow_self_connections(self):
        return self.__allow_self_connections

    @allow_self_connections.setter
    def allow_self_connections(self, new_value):
        self.__allow_self_connections = new_value

    @property
    def d_expression(self):
        return self.__d_expression

    @d_expression.setter
    def d_expression(self, new_value):
        self.__d_expression = new_value



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import numpy
from spinn_utilities.overrides import overrides
from data_specification.enums import DataType
from pacman.executor.injection_decorator import inject_items
from .abstract_synapse_type import AbstractSynapseType

TAU_SYN_E = 'tau_syn_E'
TAU_SYN_I = 'tau_syn_I'
ISYN_EXC = "isyn_exc"
ISYN_INH = "isyn_inh"

UNITS = {
    TAU_SYN_E: "mV",
    TAU_SYN_I: 'mV',
    ISYN_EXC: "",
    ISYN_INH: "",
}


class SynapseTypeExponential(AbstractSynapseType):
    __slots__ = [
        "__tau_syn_E",
        "__tau_syn_I",
        "__isyn_exc",
        "__isyn_inh"]

    def __init__(self, tau_syn_E, tau_syn_I, isyn_exc, isyn_inh):
        super(SynapseTypeExponential, self).__init__([
            DataType.U032,    # decay_E
            DataType.U032,    # init_E
            DataType.S1615,   # isyn_exc
            DataType.U032,    # decay_I
            DataType.U032,    # init_I
            DataType.S1615])  # isyn_inh
        self.__tau_syn_E = tau_syn_E
        self.__tau_syn_I = tau_syn_I
        self.__isyn_exc = isyn_exc
        self.__isyn_inh = isyn_inh

    @overrides(AbstractSynapseType.get_n_cpu_cycles)
    def get_n_cpu_cycles(self, n_neurons):
        return 100 * n_neurons

    @overrides(AbstractSynapseType.add_parameters)
    def add_parameters(self, parameters):
        parameters[TAU_SYN_E] = self.__tau_syn_E
        parameters[TAU_SYN_I] = self.__tau_syn_I

    @overrides(AbstractSynapseType.add_state_variables)
    def add_state_variables(self, state_variables):
        state_variables[ISYN_EXC] = self.__isyn_exc
        state_variables[ISYN_INH] = self.__isyn_inh

    @overrides(AbstractSynapseType.get_units)
    def get_units(self, variable):
        return UNITS[variable]

    @overrides(AbstractSynapseType.has_variable)
    def has_variable(self, variable):
        return variable in UNITS

    @inject_items({"ts": "MachineTimeStep"})
    @overrides(AbstractSynapseType.get_values, additional_arguments={'ts'})
    def get_values(self, parameters, state_variables, vertex_slice, ts):

        tsfloat = float(ts) / 1000.0
        decay = lambda x: numpy.exp(-tsfloat / x)  # noqa E731
        init = lambda x: (x / tsfloat) * (1.0 - numpy.exp(-tsfloat / x))  # noqa E731

        # Add the rest of the data
        return [parameters[TAU_SYN_E].apply_operation(decay),
                parameters[TAU_SYN_E].apply_operation(init),
                state_variables[ISYN_EXC],
                parameters[TAU_SYN_I].apply_operation(decay),
                parameters[TAU_SYN_I].apply_operation(init),
                state_variables[ISYN_INH]]

    @overrides(AbstractSynapseType.update_values)
    def update_values(self, values, parameters, state_variables):

        # Read the data
        (_decay_E, _init_E, isyn_exc, _decay_I, _init_I, isyn_inh) = values

        state_variables[ISYN_EXC] = isyn_exc
        state_variables[ISYN_INH] = isyn_inh

    @overrides(AbstractSynapseType.get_n_synapse_types)
    def get_n_synapse_types(self):
        return 2

    @overrides(AbstractSynapseType.get_synapse_id_by_target)
    def get_synapse_id_by_target(self, target):
        if target == "excitatory":
            return 0
        elif target == "inhibitory":
            return 1
        return None

    @overrides(AbstractSynapseType.get_synapse_targets)
    def get_synapse_targets(self):
        return "excitatory", "inhibitory"

    @property
    def tau_syn_E(self):
        return self.__tau_syn_E

    @tau_syn_E.setter
    def tau_syn_E(self, tau_syn_E):
        self.__tau_syn_E = tau_syn_E

    @property
    def tau_syn_I(self):
        return self.__tau_syn_I

    @tau_syn_I.setter
    def tau_syn_I(self, tau_syn_I):
        self.__tau_syn_I = tau_syn_I

    @property
    def isyn_exc(self):
        return self.__isyn_exc

    @isyn_exc.setter
    def isyn_exc(self, isyn_exc):
        self.__isyn_exc = isyn_exc

    @property
    def isyn_inh(self):
        return self.__isyn_inh

    @isyn_inh.setter
    def isyn_inh(self, isyn_inh):
        self.__isyn_inh = isyn_inh



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import division
import logging
import struct
import numpy
from spinn_front_end_common.utilities.helpful_functions import (
    locate_memory_region_for_placement)
from spynnaker.pyNN.exceptions import MemReadException

logger = logging.getLogger(__name__)
_RECORDING_COUNT = struct.Struct("<I")


def get_recording_region_size_in_bytes(
        n_machine_time_steps, bytes_per_timestep):
    """ Get the size of a recording region in bytes
    """
    if n_machine_time_steps is None:
        raise Exception(
            "Cannot record this parameter without a fixed run time")
    return ((n_machine_time_steps * bytes_per_timestep) +
            (n_machine_time_steps * 4))


def get_data(transceiver, placement, region, region_size):
    """ Get the recorded data from a region
    """

    region_base_address = locate_memory_region_for_placement(
        placement, region, transceiver)
    number_of_bytes_written = _RECORDING_COUNT.unpack_from(
        transceiver.read_memory(
            placement.x, placement.y, region_base_address,
            _RECORDING_COUNT.size))[0]

    # Subtract 4 for the word representing the size itself
    expected_size = region_size - _RECORDING_COUNT.size
    if number_of_bytes_written > expected_size:
        raise MemReadException(
            "Expected {} bytes but read {}".format(
                expected_size, number_of_bytes_written))

    return (
        transceiver.read_memory(
            placement.x, placement.y, region_base_address + 4,
            number_of_bytes_written),
        number_of_bytes_written)


def pull_off_cached_lists(no_loads, cache_file):
    """ Extracts numpy based data from a  file

    :param no_loads: the number of numpy elements in the file
    :param cache_file: the file to extract from
    :return: The extracted data
    """
    cache_file.seek(0)
    if no_loads == 1:
        values = numpy.load(cache_file)
        # Seek to the end of the file (for windows compatibility)
        cache_file.seek(0, 2)
        return values
    elif no_loads == 0:
        return []

    lists = list()
    for _ in range(0, no_loads):
        lists.append(numpy.load(cache_file))
    # Seek to the end of the file (for windows compatibility)
    cache_file.seek(0, 2)
    return numpy.concatenate(lists)


def needs_buffering(buffer_max, space_needed, enable_buffered_recording):
    if space_needed == 0:
        return False
    if not enable_buffered_recording:
        return False
    if buffer_max < space_needed:
        return True
    return False


def get_buffer_sizes(buffer_max, space_needed, enable_buffered_recording):
    if space_needed == 0:
        return 0
    if not enable_buffered_recording:
        return space_needed
    if buffer_max < space_needed:
        return buffer_max
    return space_needed


def make_missing_string(missing):
    missing_str = ""
    separator = ""
    for placement in missing:
        missing_str += "{}({}, {}, {})".format(
            separator, placement.x, placement.y, placement.p)
        separator = "; "
    return missing_str



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from spynnaker.pyNN.exceptions import SpynnakerException
from spynnaker.pyNN.models.defaults import defaults, default_initial_values


@defaults
class HHCondExp(object):
    """ Single-compartment Hodgkin-Huxley model with exponentially decaying \
        current input.
    """

    # noinspection PyPep8Naming
    @default_initial_values({"v", "gsyn_exc", "gsyn_inh"})
    def __init__(
            self, gbar_K=6.0, cm=0.2, e_rev_Na=50.0, tau_syn_E=0.2,
            tau_syn_I=2.0, i_offset=0.0, g_leak=0.01, e_rev_E=0.0,
            gbar_Na=20.0, e_rev_leak=-65.0, e_rev_I=-80, e_rev_K=-90.0,
            v_offset=-63, v=-65.0, gsyn_exc=0.0, gsyn_inh=0.0):
        # pylint: disable=too-many-arguments, too-many-locals, unused-argument
        raise SpynnakerException(
            "This neuron model is currently not supported by the tool chain")



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import collections
import numpy as np
from six import itervalues
from data_specification.enums.data_type import DataType
from spynnaker.pyNN.models.neural_projections import ProjectionApplicationEdge
from spynnaker.pyNN.models.neural_projections import ProjectionMachineEdge
from .abstract_synapse_dynamics_structural import (
    AbstractSynapseDynamicsStructural)
from spynnaker.pyNN.utilities import constants


class SynapseDynamicsStructuralCommon(AbstractSynapseDynamicsStructural):
    """ Common class that enables synaptic rewiring. It acts as a wrapper\
        around SynapseDynamicsStatic or SynapseDynamicsSTDP.\
        This means rewiring can operate in parallel with these\
        types of synapses.

        Written by Petrut Bogdan.

    :param f_rew: Frequency of rewiring (Hz). How many rewiring attempts will\
        be done per second.
    :type f_rew: int
    :param weight: Initial weight assigned to a newly formed connection
    :type weight: float
    :param delay: Delay assigned to a newly formed connection
    :type delay: int
    :param s_max: Maximum fan-in per target layer neuron
    :type s_max: int
    :param sigma_form_forward: Spread of feed-forward formation receptive field
    :type sigma_form_forward: float
    :param sigma_form_lateral: Spread of lateral formation receptive field
    :type sigma_form_lateral: float
    :param p_form_forward: Peak probability for feed-forward formation
    :type p_form_forward: float
    :param p_form_lateral: Peak probability for lateral formation
    :type p_form_lateral: float
    :param p_elim_pot: Probability of elimination of a potentiated synapse
    :type p_elim_pot: float
    :param p_elim_dep: Probability of elimination of a depressed synapse
    :type p_elim_dep: float
    :param grid: Grid shape
    :type grid: 2d int array
    :param lateral_inhibition: Flag whether to mark synapses formed within a\
        layer as inhibitory or excitatory
    :type lateral_inhibition: bool
    :param random_partner: \
        Flag whether to randomly select pre-synaptic partner for formation
    :type random_partner: bool
    :param seed: seed the random number generators
    :type seed: int
    """
    __slots__ = [
        # Frequency of rewiring (Hz)
        "__f_rew",
        # Period of rewiring (ms)
        "__p_rew",
        # Initial weight assigned to a newly formed connection
        "__initial_weight",
        # Delay assigned to a newly formed connection
        "__initial_delay",
        # Maximum fan-in per target layer neuron
        "__s_max",
        # Flag whether to mark synapses formed within a layer as
        # inhibitory or excitatory
        "__lateral_inhibition",
        # Spread of feed-forward formation receptive field
        "__sigma_form_forward",
        # Spread of lateral formation receptive field
        "__sigma_form_lateral",
        # Peak probability for feed-forward formation
        "__p_form_forward",
        # Peak probability for lateral formation
        "__p_form_lateral",
        # Probability of elimination of a depressed synapse
        "__p_elim_dep",
        # Probability of elimination of a potentiated synapse
        "__p_elim_pot",
        # Grid shape
        "__grid",
        # Flag whether to randomly select pre-synaptic partner for formation
        "__random_partner",
        # Holds initial connectivity as defined via connector
        "__connections",
        # SDRAM usage estimates are not perfect. This value adjusts estimates
        "__fudge_factor",
        # Maximum synaptic row length based on connectivity + padding
        "__actual_row_max_length",
        # The actual type of weights: static through the simulation or those
        # that can be change through STDP
        "__weight_dynamics",
        # Shared RNG seed to be written on all cores
        "__seeds",
        # Stores the actual SDRAM usage (value obtained only after writing spec
        # is finished)
        "__actual_sdram_usage",
        # Exponentially decayed probability LUT for feed-forward formations
        "__ff_distance_probabilities",
        # Exponentially decayed probability LUT for lateral formations
        "__lat_distance_probabilities"]

    default_parameters = {
        'stdp_model': None, 'f_rew': 10 ** 4, 'weight': 0, 'delay': 1,
        's_max': 32, 'sigma_form_forward': 2.5, 'sigma_form_lateral': 1,
        'p_form_forward': 0.16, 'p_form_lateral': 1.,
        'p_elim_pot': 1.36 * 10 ** -4, 'p_elim_dep': 0.0245,
        'grid': np.array([16, 16]), 'lateral_inhibition': 0,
        'random_partner': False}

    def __init__(self,
                 stdp_model=default_parameters['stdp_model'],
                 f_rew=default_parameters['f_rew'],
                 weight=default_parameters['weight'],
                 delay=default_parameters['delay'],
                 s_max=default_parameters['s_max'],
                 sigma_form_forward=default_parameters['sigma_form_forward'],
                 sigma_form_lateral=default_parameters['sigma_form_lateral'],
                 p_form_forward=default_parameters['p_form_forward'],
                 p_form_lateral=default_parameters['p_form_lateral'],
                 p_elim_dep=default_parameters['p_elim_dep'],
                 p_elim_pot=default_parameters['p_elim_pot'],
                 grid=default_parameters['grid'],
                 lateral_inhibition=default_parameters['lateral_inhibition'],
                 random_partner=default_parameters['random_partner'],
                 seed=None):
        self.__f_rew = f_rew
        self.__p_rew = 1. / self.__f_rew
        self.__initial_weight = weight
        self.__initial_delay = delay
        self.__s_max = s_max
        self.__lateral_inhibition = lateral_inhibition
        self.__sigma_form_forward = sigma_form_forward
        self.__sigma_form_lateral = sigma_form_lateral
        self.__p_form_forward = p_form_forward
        self.__p_form_lateral = p_form_lateral
        self.__p_elim_dep = p_elim_dep
        self.__p_elim_pot = p_elim_pot
        self.__grid = np.asarray(grid, dtype=int)
        self.__random_partner = random_partner
        self.__connections = {}

        self.__fudge_factor = 1.5
        self.__actual_row_max_length = self.__s_max

        self.__weight_dynamics = stdp_model

        # Generate a seed for the RNG on chip that should be the same for all
        # of the cores that have my learning rule
        _rng = np.random.RandomState(seed)
        self.__seeds = [_rng.randint(0x7FFFFFFF) for _ in range(4)]

        # Addition information -- used for SDRAM usage
        self.__actual_sdram_usage = {}

        self.__ff_distance_probabilities = \
            self.generate_distance_probability_array(
                self.__p_form_forward, self.__sigma_form_forward)
        self.__lat_distance_probabilities = \
            self.generate_distance_probability_array(
                self.__p_form_lateral, self.__sigma_form_lateral)

    @property
    def weight_dynamics(self):
        return self.__weight_dynamics

    def get_parameter_names(self):
        names = ['initial_weight', 'initial_delay', 'f_rew', 's_max',
                 'lateral_inhibition',
                 'sigma_form_forward', 'sigma_form_lateral', 'p_form_forward',
                 'p_form_lateral', 'p_elim_dep', 'p_elim_pot', 'grid',
                 'random_partner']
        return names

    def distance(self, x0, x1, grid=np.asarray([16, 16]),
                 type='euclidian'):  # @ReservedAssignment
        """ Compute the distance between points x0 and x1 place on the grid\
            using periodic boundary conditions.

        :param x0: first point in space
        :type x0: np.ndarray of ints
        :param x1: second point in space
        :type x1: np.ndarray of ints
        :param grid: shape of grid
        :type grid: np.ndarray of ints
        :param type: distance metric, i.e. euclidian or manhattan
        :type type: str
        :return: the distance
        :rtype: float
        """
        x0 = np.asarray(x0)
        x1 = np.asarray(x1)
        delta = np.abs(x0 - x1)
        if (delta[0] > grid[0] * .5) and grid[0] > 0:
            delta[0] -= grid[0]

        if (delta[1] > grid[1] * .5) and grid[1] > 0:
            delta[1] -= grid[1]

        if type == 'manhattan':
            return np.abs(delta).sum(axis=-1)
        return np.sqrt((delta ** 2).sum(axis=-1))

    def generate_distance_probability_array(self, probability, sigma):
        """ Generate the exponentially decaying probability LUTs.

        :param probability: peak probability
        :type probability: float
        :param sigma: spread
        :type sigma: float
        :return: distance-dependent probabilities
        :rtype: numpy.ndarray(float)
        """
        euclidian_distances = np.ones(self.__grid ** 2) * np.nan
        for row in range(euclidian_distances.shape[0]):
            for column in range(euclidian_distances.shape[1]):
                if self.__grid[0] > 1:
                    pre = (row // self.__grid[0], row % self.__grid[1])
                    post = (column // self.__grid[0], column % self.__grid[1])
                else:
                    pre = (0, row % self.__grid[1])
                    post = (0, column % self.__grid[1])

                euclidian_distances[row, column] = self.distance(
                    pre, post, grid=self.__grid, type='euclidian')
        largest_squared_distance = np.max(euclidian_distances ** 2)
        squared_distances = np.arange(largest_squared_distance)
        raw_probabilities = probability * (
            np.exp(-squared_distances / (2 * sigma ** 2)))
        quantised_probabilities = raw_probabilities * ((2 ** 16) - 1)
        # Quantize probabilities and cast as uint16 / short
        unfiltered_probabilities = quantised_probabilities.astype(
            dtype="uint16")
        # Only return probabilities which are non-zero
        filtered_probabilities = unfiltered_probabilities[
            unfiltered_probabilities > 0]
        if filtered_probabilities.size % 2 != 0:
            filtered_probabilities = np.concatenate(
                (filtered_probabilities,
                 np.zeros(filtered_probabilities.size % 2, dtype="uint16")))

        return filtered_probabilities

    @property
    def p_rew(self):
        """ The period of rewiring.

        :return: The period of rewiring
        :rtype: int
        """
        return self.__p_rew

    @property
    def actual_sdram_usage(self):
        """ Actual SDRAM usage (based on what is written to spec).

        :return: actual SDRAM usage
        :rtype: int
        """
        return self.__actual_sdram_usage

    def write_parameters(
            self, spec, region, machine_time_step, weight_scales,
            application_graph, machine_graph, app_vertex, post_slice,
            machine_vertex, graph_mapper, routing_info):
        """ Write the synapse parameters to the spec.

        :param spec: the data spec
        :type spec: spec
        :param region: memory region
        :type region: int
        :param machine_time_step: the duration of a machine time step (ms)
        :type machine_time_step: int
        :param weight_scales: scaling the weights
        :type weight_scales: list(float)
        :param application_graph: \
            the entire, highest level, graph of the network to be simulated
        :type application_graph: :py:class:`ApplicationGraph`
        :param machine_graph: \
            the entire, lowest level, graph of the network to be simulated
        :type machine_graph: :py:class:`MachineGraph`
        :param app_vertex: \
            the highest level object of the post-synaptic population
        :type app_vertex: :py:class:`ApplicationVertex`
        :param post_slice: \
            the slice of the app vertex corresponding to this machine vertex
        :type post_slice: :py:class:`Slice`
        :param machine_vertex: \
            the lowest level object of the post-synaptic population
        :type machine_vertex: :py:class:`MachineVertex`
        :param graph_mapper: for looking up application vertices
        :type graph_mapper: :py:class:`GraphMapper`
        :param routing_info: All of the routing information on the network
        :type routing_info: :py:class:`RoutingInfo`
        :return: None
        :rtype: None
        """
        spec.comment("Writing structural plasticity parameters")
        if spec.current_region != constants.POPULATION_BASED_REGIONS. \
                SYNAPSE_DYNAMICS.value:
            spec.switch_write_focus(region)

        # Write the common part of the rewiring data
        self.__write_common_rewiring_data(
            spec, app_vertex, post_slice, weight_scales, machine_time_step)

        # Write presynaptic (sub)population information
        self.__write_presynaptic_information(
            spec, application_graph, machine_graph,
            app_vertex, post_slice, machine_vertex, graph_mapper,
            routing_info)

    def __write_common_rewiring_data(self, spec, app_vertex, post_slice,
                                     weight_scales, machine_time_step):
        """ Write the non-subpopulation synapse parameters to the spec.

        :param spec: the data spec
        :type spec: spec
        :param app_vertex: \
            the highest level object of the post-synaptic population
        :type app_vertex: :py:class:`ApplicationVertex`
        :param post_slice: \
            the slice of the app vertex corresponding to this machine vertex
        :type post_slice: :py:class:`Slice`
        :param weight_scales: scaling the weights
        :type weight_scales: list(float)
        :param machine_time_step: the duration of a machine time step (ms)
        :type machine_time_step: int
        :return: None
        :rtype: None
        """
        if self.__p_rew * 1000. < machine_time_step / 1000.:
            # Fast rewiring
            spec.write_value(data=1)
            spec.write_value(
                data=int(machine_time_step / (self.__p_rew * 10 ** 6)))
        else:
            # Slow rewiring
            spec.write_value(data=0)
            spec.write_value(
                data=int((self.__p_rew * 10 ** 6) / float(machine_time_step)))

        # scale the excitatory weight appropriately
        spec.write_value(
            data=int(round(self.__initial_weight * weight_scales[0])))
        # scale the inhibitory weight appropriately
        spec.write_value(
            data=int(round(self.__initial_weight * weight_scales[1])))
        spec.write_value(data=self.__initial_delay)
        spec.write_value(data=int(self.__s_max))
        spec.write_value(data=int(self.__lateral_inhibition),
                         data_type=DataType.INT32)
        spec.write_value(data=int(self.__random_partner),
                         data_type=DataType.INT32)
        # write total number of atoms in the application vertex
        spec.write_value(data=app_vertex.n_atoms)
        # write local low, high and number of atoms
        spec.write_value(data=post_slice[0])
        spec.write_value(data=post_slice[1])
        spec.write_value(data=post_slice[2])

        # write the grid size for periodic boundary distance computation
        spec.write_value(data=self.__grid[0])
        spec.write_value(data=self.__grid[1])

        # write probabilities for elimination
        spec.write_value(data=self.__p_elim_dep, data_type=DataType.U032)
        spec.write_value(data=self.__p_elim_pot, data_type=DataType.U032)

        # write the random seed (4 words), generated randomly,
        # but the same for all postsynaptic vertices!
        for seed in self.__seeds:
            spec.write_value(data=seed)

        # write local seed (4 words), generated randomly!
        for _ in range(4):
            spec.write_value(data=np.random.randint(0x7FFFFFFF))

    def __compute_aux(self, application_graph, machine_graph,
                      app_vertex, machine_vertex, graph_mapper, routing_info):
        """ Compute all of the relevant pre-synaptic population information,\
            as well as the key of the current vertex.

        :param application_graph: \
            the entire, highest level, graph of the network to be simulated
        :type application_graph: :py:class:`ApplicationGraph`
        :param machine_graph: \
            the entire, lowest level, graph of the network to be simulated
        :type machine_graph: :py:class:`MachineGraph`
        :param app_vertex: \
            the highest level object of the post-synaptic population
        :type app_vertex: :py:class:`ApplicationVertex`
        :param post_slice: \
            the slice of the app vertex corresponding to this machine vertex
        :type post_slice: :py:class:`Slice`
        :param machine_vertex: \
            the lowest level object of the post-synaptic population
        :type machine_vertex: :py:class:`MachineVertex`
        :param graph_mapper: for looking up application vertices
        :type graph_mapper: :py:class:`GraphMapper`
        :param routing_info: All of the routing information on the network
        :type routing_info: :py:class:`RoutingInfo`
        :return: pop info, routing key for current vertex, number of pre pops
        :rtype: tuple
        """
        # Compute the max number of presynaptic subpopulations
        population_to_subpopulation_information = collections.OrderedDict()

        # Can figure out the presynaptic subvertices (machine vertices)
        # for the current machine vertex
        # by calling graph_mapper.get_machine_edges for the relevant
        # application edges (i.e. the structural ones)
        # This allows me to find the partition (?) which then plugged
        # into routing_info can give me the keys
        presynaptic_machine_vertices = []
        structural_application_edges = []
        structural_machine_edges = []
        max_subpartitions = 0

        for app_edge in application_graph.get_edges_ending_at_vertex(
                app_vertex):
            if isinstance(app_edge, ProjectionApplicationEdge):
                for synapse_info in app_edge.synapse_information:
                    if synapse_info.synapse_dynamics is self.__weight_dynamics:
                        structural_application_edges.append(app_edge)
                        population_to_subpopulation_information[
                            app_edge.pre_vertex] = []
                        break

        no_pre_populations = len(structural_application_edges)
        # For each structurally plastic APPLICATION edge find the
        # corresponding machine edges
        for machine_edge in machine_graph.get_edges_ending_at_vertex(
                machine_vertex):
            if isinstance(machine_edge, ProjectionMachineEdge):
                for synapse_info in machine_edge._synapse_information:
                    if synapse_info.synapse_dynamics is self.__weight_dynamics:
                        structural_machine_edges.append(machine_edge)
                        # For each structurally plastic MACHINE edge find the
                        # corresponding presynaptic subvertices
                        presynaptic_machine_vertices.append(
                            machine_edge.pre_vertex)

        # For each presynaptic subvertex figure out the partition (?)
        # to retrieve the key and n_atoms
        for vertex in presynaptic_machine_vertices:
            population_to_subpopulation_information[
                graph_mapper.get_application_vertex(vertex)].append(
                (routing_info.get_routing_info_from_pre_vertex(
                    vertex, constants.SPIKE_PARTITION_ID).first_key,
                 graph_mapper.get_slice(vertex)[2],
                 graph_mapper.get_slice(vertex)[0],
                 routing_info.get_routing_info_from_pre_vertex(
                     vertex, constants.SPIKE_PARTITION_ID).first_mask))

        for subpopulation_list in itervalues(
                population_to_subpopulation_information):
            max_subpartitions = np.maximum(max_subpartitions,
                                           len(subpopulation_list))

        # Current machine vertex key (for future checks)
        current_key = routing_info.get_routing_info_from_pre_vertex(
            machine_vertex, constants.SPIKE_PARTITION_ID)

        if current_key is not None:
            current_key = current_key.first_key
        else:
            current_key = -1
        return (population_to_subpopulation_information, current_key,
                no_pre_populations)

    def __write_presynaptic_information(self, spec, application_graph,
                                        machine_graph,
                                        app_vertex, post_slice, machine_vertex,
                                        graph_mapper,
                                        routing_info):
        """ All cores which do synaptic rewiring have information about all\
            the relevant pre-synaptic populations.

        :param spec: the data spec
        :type spec: spec
        :param application_graph: \
            the entire, highest level, graph of the network to be simulated
        :type application_graph: :py:class:`ApplicationGraph`
        :param machine_graph: \
            the entire, lowest level, graph of the network to be simulated
        :type machine_graph: :py:class:`MachineGraph`
        :param app_vertex: \
            the highest level object of the post-synaptic population
        :type app_vertex: :py:class:`ApplicationVertex`
        :param post_slice: \
            the slice of the app vertex corresponding to this machine vertex
        :type post_slice: :py:class:`Slice`
        :param machine_vertex: \
            the lowest level object of the post-synaptic population
        :type machine_vertex: :py:class:`MachineVertex`
        :param graph_mapper: for looking up application vertices
        :type graph_mapper: :py:class:`GraphMapper`
        :param routing_info: All of the routing information on the network
        :type routing_info: :py:class:`RoutingInfo`
        :return: None
        :rtype: None
        """
        # Compute all the auxiliary stuff
        pop_to_subpop_info, current_key, no_prepops = self.__compute_aux(
            application_graph, machine_graph, app_vertex, machine_vertex,
            graph_mapper, routing_info)

        # Table header
        spec.write_value(data=no_prepops)

        total_words_written = 0
        for subpopulation_list in itervalues(pop_to_subpop_info):
            # Population header(s)
            # Number of subpopulations
            spec.write_value(data=len(subpopulation_list),
                             data_type=DataType.UINT16)

            # Custom header for commands / controls
            # currently, controls = True if the subvertex (on the current core)
            # is part of this population
            controls = current_key in np.asarray(subpopulation_list)[:0]
            spec.write_value(data=int(controls), data_type=DataType.UINT16)

            spec.write_value(
                data=np.sum(np.asarray(subpopulation_list)[:, 1]) if len(
                    subpopulation_list) > 0 else 0)
            words_written = 2

            # Ensure the following values are written in ascending
            # order of low_atom (implicit)
            dt = np.dtype(
                [('key', 'uint'), ('n_atoms', 'uint'), ('lo_atom', 'uint'),
                 ('mask', 'uint')])
            structured_array = np.array(subpopulation_list, dtype=dt)
            sorted_info_list = np.sort(structured_array, order='lo_atom')
            for subpopulation_info in sorted_info_list:
                # Subpopulation information (i.e. key and number of atoms)
                # Key
                spec.write_value(data=subpopulation_info[0])
                # n_atoms
                spec.write_value(data=subpopulation_info[1])
                # lo_atom
                spec.write_value(data=subpopulation_info[2])
                # mask
                spec.write_value(data=subpopulation_info[3])
                words_written += 4

            total_words_written += words_written

        # Now we write the probability tables for formation
        # (feedforward and lateral)
        spec.write_value(data=self.__ff_distance_probabilities.size)
        spec.write_array(
            self.__ff_distance_probabilities.view(dtype=np.uint16),
            data_type=DataType.UINT16)
        total_words_written += self.__ff_distance_probabilities.size // 2 + 1
        spec.write_value(data=self.__lat_distance_probabilities.size,
                         data_type=DataType.INT32)
        spec.write_array(
            self.__lat_distance_probabilities.view(dtype=np.uint16),
            data_type=DataType.UINT16)
        total_words_written += self.__lat_distance_probabilities.size // 2 + 1

        # Write post to pre table (inverse of synaptic matrix)
        self.__write_post_to_pre_table(spec, app_vertex, post_slice,
                                       machine_vertex, graph_mapper,
                                       pop_to_subpop_info, total_words_written)

    def __write_post_to_pre_table(self, spec, app_vertex, post_slice,
                                  machine_vertex, graph_mapper,
                                  population_to_subpopulation_information,
                                  total_words_written):
        """ Post to pre table is basically the transpose of the synaptic\
            matrix

        :param spec: the data spec
        :type spec: spec
        :param app_vertex: \
            the highest level object of the post-synaptic population
        :type app_vertex: ApplicationVertex
        :param post_slice: \
            the slice of the app vertex corresponding to this machine vertex
        :type post_slice: Slice
        :param machine_vertex: \
            the lowest level object of the post-synaptic population
        :type machine_vertex: MachineVertex
        :param graph_mapper: for looking up application vertices
        :type graph_mapper: GraphMapper
        :param population_to_subpopulation_information: \
            generated relevant information
        :type population_to_subpopulation_information: dict
        :param total_words_written: \
            keeping track of how many words have been written
        :type total_words_written: int
        :return: None
        :rtype: None
        """
        # Setting up Post to Pre table
        post_to_pre_table = np.ones((post_slice.n_atoms, self.__s_max),
                                    dtype=np.int32) * -1
        for row in self.__connections[post_slice.lo_atom]:
            if row[0].size > 0 and row[1].post_vertex is app_vertex:
                for source, target, _weight, _delay, _syn_type in row[0]:

                    # Select pre vertex
                    pre_vertex_slice = graph_mapper._slice_by_machine_vertex[
                        row[2].pre_vertex]
                    pre_vertex_id = source - pre_vertex_slice.lo_atom
                    masked_pre_vertex_id = pre_vertex_id & (2 ** 17 - 1)

                    # Select population index
                    pop_index = population_to_subpopulation_information. \
                        keys().index(row[1].pre_vertex)
                    masked_pop_index = pop_index & (2 ** 9 - 1)

                    # Select subpopulation index
                    dt = np.dtype(
                        [('key', 'int'), ('n_atoms', 'int'),
                         ('lo_atom', 'int'), ('mask', 'uint')])
                    structured_array = np.array(
                        population_to_subpopulation_information[
                            row[1].pre_vertex], dtype=dt)
                    sorted_info_list = np.sort(
                        structured_array, order='lo_atom')

                    # find index where lo_atom equals the one in
                    # pre_vertex_slice
                    subpop_index = np.argwhere(
                        sorted_info_list['lo_atom'] ==
                        pre_vertex_slice.lo_atom).ravel()[0]
                    masked_sub_pop_index = subpop_index & (2 ** 9 - 1)

                    # identifier combines the vertex, pop and subpop
                    # into 1 x 32 bit word
                    identifier = (masked_pop_index << (32 - 8)) | (
                            masked_sub_pop_index << 16) | masked_pre_vertex_id
                    try:
                        synaptic_entry = np.argmax(
                            post_to_pre_table[target - post_slice.lo_atom] ==
                            -1).ravel()[0]
                    except Exception:
                        break
                    post_to_pre_table[
                        target - post_slice.lo_atom, synaptic_entry] = \
                        identifier

        spec.write_array(post_to_pre_table.ravel(), data_type=DataType.INT32)
        total_words_written += post_to_pre_table.size

        self.actual_sdram_usage[
            machine_vertex] = 4 * 27 + 4 * total_words_written

    def get_extra_sdram_usage_in_bytes(self, machine_in_edges):
        """ Better approximation of SDRAM usage based on incoming machine edges

        :param machine_in_edges: incoming machine edges
        :type machine_in_edges: machine edges
        :return: SDRAM usage
        :rtype: int
        """
        relevant_edges = []
        for edge in machine_in_edges:
            for synapse_info in edge._synapse_information:
                if synapse_info.synapse_dynamics is self.__weight_dynamics:
                    relevant_edges.append(edge)
        return int(self.__fudge_factor * 4 * 12 * len(relevant_edges))

    def get_parameters_sdram_usage_in_bytes(self, n_neurons, n_synapse_types,
                                            in_edges):
        """ Approximate SDRAM usage

        :param n_neurons: number of neurons
        :type n_neurons: int
        :param n_synapse_types: \
            number of synapse types (i.e. excitatory and inhibitory)
        :type n_synapse_types: int
        :param in_edges: incoming edges
        :type in_edges: edges
        :return: SDRAM usage
        :rtype: int
        """
        structure_size = 27 * 4 + 4 * 4  # parameters + rng seed
        post_to_pre_table_size = n_neurons * self.__s_max * 4
        structure_size += post_to_pre_table_size

        initial_size = 0
        total_size = structure_size + initial_size

        # Approximation of the sizes of both probability vs distance tables
        ff_lut = np.count_nonzero(self.__ff_distance_probabilities) * 4
        lat_lut = np.count_nonzero(self.__lat_distance_probabilities) * 4
        total_size += ff_lut
        total_size += lat_lut
        # total_size += (80 * 4)
        pop_size = 0

        # approximate the size of the pop -> subpop table
        no_pre_vertices_estimate = 0
        if in_edges is not None:

            # Approximation gets computed here based on number of
            # afferent edges
            # How many afferent application vertices?
            # How large are they?

            for edge in in_edges:
                if isinstance(edge, ProjectionApplicationEdge):
                    for synapse_info in edge.synapse_information:
                        if (synapse_info.synapse_dynamics is
                                self.__weight_dynamics):
                            no_pre_vertices_estimate += (1 + np.ceil(
                                edge.pre_vertex.n_atoms / 32.))
                    no_pre_vertices_estimate *= 2
                elif isinstance(edge, ProjectionMachineEdge):
                    pop_size += self.get_extra_sdram_usage_in_bytes(in_edges)
                    break

        pop_size += int(50 * (no_pre_vertices_estimate + len(in_edges)))

        return int(self.__fudge_factor * (total_size + pop_size))  # bytes

    def synaptic_data_update(self, connections,
                             post_vertex_slice,
                             app_edge, machine_edge):
        """ Get static synaptic data
        """
        if post_vertex_slice.lo_atom not in self.__connections.keys():
            self.__connections[post_vertex_slice.lo_atom] = []
        self.__connections[post_vertex_slice.lo_atom].append(
            (connections, app_edge, machine_edge))

    def n_words_for_plastic_connections(self, value):
        """ Get size of plastic connections in words
        """
        self.__actual_row_max_length = value

    def n_words_for_static_connections(self, value):
        """ Get size of static connections in words
        """
        self.__actual_row_max_length = value

    def get_n_synapses_in_rows(self, pp_size, fp_size=None):
        """ Get number of synapses in a row.
        """
        if fp_size is not None:
            return self.__weight_dynamics. \
                get_n_synapses_in_rows(pp_size, fp_size)
        return self.__weight_dynamics.get_n_synapses_in_rows(pp_size)

    def get_vertex_executable_suffix(self):
        name = "_structural"
        return name

    def is_same_as(self, synapse_dynamics):
        if not isinstance(synapse_dynamics, AbstractSynapseDynamicsStructural):
            return False
        return (
                self.__f_rew == synapse_dynamics.f_rew and
                self.__s_max == synapse_dynamics.s_max and
                np.isclose(self.__sigma_form_forward,
                           synapse_dynamics.sigma_form_forward) and
                np.isclose(self.__sigma_form_lateral,
                           synapse_dynamics.sigma_form_lateral) and
                np.isclose(self.__p_form_forward,
                           synapse_dynamics.p_form_forward) and
                np.isclose(self.__p_form_lateral,
                           synapse_dynamics.p_form_lateral) and
                np.isclose(self.__p_elim_dep, synapse_dynamics.p_elim_dep) and
                np.isclose(self.__p_elim_pot, synapse_dynamics.p_elim_pot))



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import math
import logging
from data_specification.enums import DataType
from spinn_front_end_common.utilities.utility_objs import ProvenanceDataItem

logger = logging.getLogger(__name__)
# Default value of fixed-point one for STDP
STDP_FIXED_POINT_ONE = (1 << 11)


def float_to_fixed(value, fixed_point_one):
    return int(round(float(value) * float(fixed_point_one)))


def write_exp_lut(spec, time_constant, size, shift,
                  fixed_point_one=STDP_FIXED_POINT_ONE):
    # Calculate time constant reciprocal
    time_constant_reciprocal = 1.0 / float(time_constant)

    # Generate LUT
    last_value = None
    for i in range(size):

        # Apply shift to get time from index
        time = (i << shift)

        # Multiply by time constant and calculate negative exponential
        value = float(time) * time_constant_reciprocal
        exp_float = math.exp(-value)

        # Convert to fixed-point and write to spec
        last_value = float_to_fixed(exp_float, fixed_point_one)
        spec.write_value(data=last_value, data_type=DataType.INT16)

    # return last value reverted to float (should be 0 if correct)
    return float(last_value) / float(fixed_point_one)


def get_lut_provenance(
        pre_population_label, post_population_label, rule_name, entry_name,
        param_name, last_entry):
    # pylint: disable=too-many-arguments
    top_level_name = "{}_{}_STDP_{}".format(
        pre_population_label, post_population_label, rule_name)
    report = False
    if last_entry is not None:
        report = last_entry > 0
    return ProvenanceDataItem(
        [top_level_name, entry_name], last_entry, report=report,
        message=(
            "The last entry in the STDP exponential lookup table for the {}"
            " parameter of the {} between {} and {} was {} rather than 0,"
            " indicating that the lookup table was not big enough at this"
            " timestep and value.  Try reducing the parameter value, or"
            " increasing the timestep".format(
                param_name, rule_name, pre_population_label,
                post_population_label, last_entry)))



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import division
try:
    from collections.abc import OrderedDict
except ImportError:
    from collections import OrderedDict
import logging
import math
import numpy
from six import raise_from, iteritems
from six.moves import range, xrange
from spinn_utilities.index_is_value import IndexIsValue
from spinn_utilities.progress_bar import ProgressBar
from pacman.model.resources.variable_sdram import VariableSDRAM
from data_specification.enums import DataType
from spinn_front_end_common.utilities.exceptions import ConfigurationException
from spinn_front_end_common.utilities import globals_variables
from spynnaker.pyNN.models.neural_properties import NeuronParameter

logger = logging.getLogger(__name__)
SPIKES = "spikes"


class _ReadOnlyDict(dict):
    def __readonly__(self, *args, **kwargs):
        raise RuntimeError("Cannot modify ReadOnlyDict")

    __setitem__ = __readonly__
    __delitem__ = __readonly__
    pop = __readonly__
    popitem = __readonly__
    clear = __readonly__
    update = __readonly__
    setdefault = __readonly__
    del __readonly__


class NeuronRecorder(object):
    __slots__ = [
        "__indexes", "__n_neurons", "__sampling_rates"]

    N_BYTES_FOR_TIMESTAMP = 4
    N_BYTES_PER_VALUE = 4
    N_BYTES_PER_RATE = 4  # uint32
    N_BYTES_PER_INDEX = 1  # currently uint8
    N_BYTES_PER_SIZE = 4
    N_CPU_CYCLES_PER_NEURON = 8
    N_BYTES_PER_WORD = 4
    N_BYTES_PER_POINTER = 4
    SARK_BLOCK_SIZE = 8  # Seen in sark.c

    MAX_RATE = 2 ** 32 - 1  # To allow a unit32_t to be used to store the rate

    def __init__(self, allowed_variables, n_neurons):
        self.__sampling_rates = OrderedDict()
        self.__indexes = dict()
        self.__n_neurons = n_neurons
        for variable in allowed_variables:
            self.__sampling_rates[variable] = 0
            self.__indexes[variable] = None

    def _count_recording_per_slice(self, variable, vertex_slice):
        if self.__sampling_rates[variable] == 0:
            return 0
        if self.__indexes[variable] is None:
            return vertex_slice.n_atoms
        return sum(vertex_slice.lo_atom <= index <= vertex_slice.hi_atom
                   for index in self.__indexes[variable])

    def _neurons_recording(self, variable, vertex_slice):
        if self.__sampling_rates[variable] == 0:
            return []
        if self.__indexes[variable] is None:
            return range(vertex_slice.lo_atom, vertex_slice.hi_atom+1)
        recording = []
        indexes = self.__indexes[variable]
        for index in xrange(vertex_slice.lo_atom, vertex_slice.hi_atom+1):
            if index in indexes:
                recording.append(index)
        return recording

    def get_neuron_sampling_interval(self, variable):
        """ Return the current sampling interval for this variable

        :param variable: PyNN name of the variable
        :return: Sampling interval in micro seconds
        """
        step = globals_variables.get_simulator().machine_time_step / 1000
        return self.__sampling_rates[variable] * step

    def get_matrix_data(
            self, label, buffer_manager, region, placements, graph_mapper,
            application_vertex, variable, n_machine_time_steps):
        """ Read a uint32 mapped to time and neuron IDs from the SpiNNaker\
            machine.

        :param label: vertex label
        :param buffer_manager: the manager for buffered data
        :param region: the DSG region ID used for this data
        :param placements: the placements object
        :param graph_mapper: \
            the mapping between application and machine vertices
        :param application_vertex:
        :param variable: PyNN name for the variable (V, gsy_inh etc.)
        :type variable: str
        :param n_machine_time_steps:
        :return:
        """
        if variable == SPIKES:
            msg = "Variable {} is not supported use get_spikes".format(SPIKES)
            raise ConfigurationException(msg)
        vertices = graph_mapper.get_machine_vertices(application_vertex)
        progress = ProgressBar(
            vertices, "Getting {} for {}".format(variable, label))
        sampling_rate = self.__sampling_rates[variable]
        expected_rows = int(math.ceil(
            n_machine_time_steps / sampling_rate))
        missing_str = ""
        data = None
        indexes = []
        for vertex in progress.over(vertices):
            placement = placements.get_placement_of_vertex(vertex)
            vertex_slice = graph_mapper.get_slice(vertex)
            neurons = self._neurons_recording(variable, vertex_slice)
            n_neurons = len(neurons)
            if n_neurons == 0:
                continue
            indexes.extend(neurons)
            # for buffering output info is taken form the buffer manager
            record_raw, missing_data = buffer_manager.get_data_by_placement(
                    placement, region)
            record_length = len(record_raw)

            row_length = self.N_BYTES_FOR_TIMESTAMP + \
                n_neurons * self.N_BYTES_PER_VALUE

            # There is one column for time and one for each neuron recording
            n_rows = record_length // row_length
            if record_length > 0:
                # Converts bytes to ints and make a matrix
                record = (numpy.asarray(record_raw, dtype="uint8").
                          view(dtype="<i4")).reshape((n_rows, (n_neurons + 1)))
            else:
                record = numpy.empty((0, n_neurons))
            # Check if you have the expected data
            if not missing_data and n_rows == expected_rows:
                # Just cut the timestamps off to get the fragment
                fragment = (record[:, 1:] / float(DataType.S1615.scale))
            else:
                missing_str += "({}, {}, {}); ".format(
                    placement.x, placement.y, placement.p)
                # Start the fragment for this slice empty
                fragment = numpy.empty((expected_rows, n_neurons))
                for i in xrange(0, expected_rows):
                    time = i * sampling_rate
                    # Check if there is data for this timestep
                    local_indexes = numpy.where(record[:, 0] == time)
                    if len(local_indexes[0]) == 1:
                        fragment[i] = (record[local_indexes[0], 1:] /
                                       float(DataType.S1615.scale))
                    elif len(local_indexes[0]) > 1:
                        logger.warning(
                            "Population {} on multiple recorded data for "
                            "time {}".format(label, time))
                    else:
                        # Set row to nan
                        fragment[i] = numpy.full(n_neurons, numpy.nan)
            if data is None:
                data = fragment
            else:
                # Add the slice fragment on axis 1 which is IDs/channel_index
                data = numpy.append(data, fragment, axis=1)
        if len(missing_str) > 0:
            logger.warning(
                "Population {} is missing recorded data in region {} from the"
                " following cores: {}".format(label, region, missing_str))
        sampling_interval = self.get_neuron_sampling_interval(variable)
        return (data, indexes, sampling_interval)

    def get_spikes(
            self, label, buffer_manager, region, placements, graph_mapper,
            application_vertex, machine_time_step):

        spike_times = list()
        spike_ids = list()
        ms_per_tick = machine_time_step / 1000.0

        vertices = graph_mapper.get_machine_vertices(application_vertex)
        missing_str = ""
        progress = ProgressBar(vertices,
                               "Getting spikes for {}".format(label))
        for vertex in progress.over(vertices):
            placement = placements.get_placement_of_vertex(vertex)
            vertex_slice = graph_mapper.get_slice(vertex)

            if self.__indexes[SPIKES] is None:
                neurons_recording = vertex_slice.n_atoms
            else:
                neurons_recording = sum(
                    (index >= vertex_slice.lo_atom and
                     index <= vertex_slice.hi_atom)
                    for index in self.__indexes[SPIKES])
                if neurons_recording == 0:
                    continue
            # Read the spikes
            n_words = int(math.ceil(neurons_recording / 32.0))
            n_bytes = n_words * self.N_BYTES_PER_WORD
            n_words_with_timestamp = n_words + 1

            # for buffering output info is taken form the buffer manager
            record_raw, data_missing = buffer_manager.get_data_by_placement(
                    placement, region)
            if data_missing:
                missing_str += "({}, {}, {}); ".format(
                    placement.x, placement.y, placement.p)
            if len(record_raw) > 0:
                raw_data = (numpy.asarray(record_raw, dtype="uint8").
                            view(dtype="<i4")).reshape(
                    [-1, n_words_with_timestamp])
            else:
                raw_data = record_raw
            if len(raw_data) > 0:
                record_time = raw_data[:, 0] * float(ms_per_tick)
                spikes = raw_data[:, 1:].byteswap().view("uint8")
                bits = numpy.fliplr(numpy.unpackbits(spikes).reshape(
                    (-1, 32))).reshape((-1, n_bytes * 8))
                time_indices, local_indices = numpy.where(bits == 1)
                if self.__indexes[SPIKES] is None:
                    indices = local_indices + vertex_slice.lo_atom
                    times = record_time[time_indices].reshape((-1))
                    spike_ids.extend(indices)
                    spike_times.extend(times)
                else:
                    neurons = self._neurons_recording(SPIKES, vertex_slice)
                    n_neurons = len(neurons)
                    for time_indice, local in zip(time_indices, local_indices):
                        if local < n_neurons:
                            spike_ids.append(neurons[local])
                            spike_times.append(record_time[time_indice])

        if len(missing_str) > 0:
            logger.warning(
                "Population {} is missing spike data in region {} from the"
                " following cores: {}".format(label, region, missing_str))

        if len(spike_ids) == 0:
            return numpy.zeros((0, 2), dtype="float")

        result = numpy.column_stack((spike_ids, spike_times))
        return result[numpy.lexsort((spike_times, spike_ids))]

    def get_recordable_variables(self):
        return self.__sampling_rates.keys()

    def is_recording(self, variable):
        try:
            return self.__sampling_rates[variable] > 0
        except KeyError as e:
            msg = "Variable {} is not supported. Supported variables are {}" \
                  "".format(variable, self.get_recordable_variables())
            raise_from(ConfigurationException(msg), e)

    @property
    def recording_variables(self):
        results = list()
        for region, rate in self.__sampling_rates.items():
            if rate > 0:
                results.append(region)
        return results

    @property
    def recorded_region_ids(self):
        results = list()
        for id, rate in enumerate(self.__sampling_rates.values()):
            if rate > 0:
                results.append(id)
        return results

    def _compute_rate(self, sampling_interval):
        """ Convert a sampling interval into a rate. \
            Remember, machine time step is in nanoseconds

        :param sampling_interval: interval between samples in microseconds
        :return: rate
        """
        if sampling_interval is None:
            return 1

        step = globals_variables.get_simulator().machine_time_step / 1000
        rate = int(sampling_interval / step)
        if sampling_interval != rate * step:
            msg = "sampling_interval {} is not an an integer multiple of the "\
                  "simulation timestep {}".format(sampling_interval, step)
            raise ConfigurationException(msg)
        if rate > self.MAX_RATE:
            msg = "sampling_interval {} higher than max allowed which is {}" \
                  "".format(sampling_interval, step * self.MAX_RATE)
            raise ConfigurationException(msg)
        return rate

    def check_indexes(self, indexes):
        if indexes is None:
            return

        if len(indexes) == 0:
            raise ConfigurationException("Empty indexes list")

        found = False
        warning = None
        for index in indexes:
            if index < 0:
                raise ConfigurationException(
                    "Negative indexes are not supported")
            elif index >= self.__n_neurons:
                warning = "Ignoring indexes greater than population size."
            else:
                found = True
            if warning is not None:
                logger.warning(warning)
        if not found:
            raise ConfigurationException(
                "All indexes larger than population size")

    def _turn_off_recording(self, variable, sampling_interval, remove_indexes):
        if self.__sampling_rates[variable] == 0:
            # Already off so ignore other parameters
            return

        if remove_indexes is None:
            # turning all off so ignoring sampling interval
            self.__sampling_rates[variable] = 0
            self.__indexes[variable] = None
            return

        # No good reason to specify_interval when turning off
        if sampling_interval is not None:
            rate = self._compute_rate(sampling_interval)
            # But if they do make sure it is the same as before
            if rate != self.__sampling_rates[variable]:
                raise ConfigurationException(
                    "Illegal sampling_interval parameter while turning "
                    "off recording")

        if self.__indexes[variable] is None:
            # start with all indexes
            self.__indexes[variable] = range(self.__n_neurons)

        # remove the indexes not recording
        self.__indexes[variable] = \
            [index for index in self.__indexes[variable]
                if index not in remove_indexes]

        # Check is at least one index still recording
        if len(self.__indexes[variable]) == 0:
            self.__sampling_rates[variable] = 0
            self.__indexes[variable] = None

    def _check_complete_overwrite(self, variable, indexes):
        if indexes is None:
            # overwriting all OK!
            return
        if self.__indexes[variable] is None:
            if set(set(range(self.__n_neurons))).issubset(set(indexes)):
                # overwriting all previous so OK!
                return
        else:
            if set(self.__indexes[variable]).issubset(set(indexes)):
                # overwriting all previous so OK!
                return
        raise ConfigurationException(
            "Current implementation does not support multiple "
            "sampling_intervals for {} on one population.".format(
                variable))

    def _turn_on_recording(self, variable, sampling_interval, indexes):

        rate = self._compute_rate(sampling_interval)
        if self.__sampling_rates[variable] == 0:
            # Previously not recording so OK
            self.__sampling_rates[variable] = rate
        elif rate != self.__sampling_rates[variable]:
            self._check_complete_overwrite(variable, indexes)
        # else rate not changed so no action

        if indexes is None:
            # previous recording indexes does not matter as now all (None)
            self.__indexes[variable] = None
        else:
            # make sure indexes is not a generator like range
            indexes = list(indexes)
            self.check_indexes(indexes)
            if self.__indexes[variable] is not None:
                # merge the two indexes
                indexes = self.__indexes[variable] + indexes
            # Avoid duplicates and keep in numerical order
            self.__indexes[variable] = list(set(indexes))
            self.__indexes[variable].sort()

    def set_recording(self, variable, new_state, sampling_interval=None,
                      indexes=None):
        if variable == "all":
            for key in self.__sampling_rates.keys():
                self.set_recording(key, new_state, sampling_interval, indexes)
        elif variable in self.__sampling_rates:
            if new_state:
                self._turn_on_recording(variable, sampling_interval, indexes)
            else:
                self._turn_off_recording(variable, sampling_interval, indexes)
        else:
            raise ConfigurationException("Variable {} is not supported".format(
                variable))

    def get_buffered_sdram_per_record(self, variable, vertex_slice):
        """ Return the SDRAM used per record

        :param variable:
        :param vertex_slice:
        :return:
        """
        n_neurons = self._count_recording_per_slice(variable, vertex_slice)
        if n_neurons == 0:
            return 0
        if variable == SPIKES:
            # Overflow can be ignored as it is not save if in an extra word
            out_spike_words = int(math.ceil(n_neurons / 32.0))
            out_spike_bytes = out_spike_words * self.N_BYTES_PER_WORD
            return self.N_BYTES_FOR_TIMESTAMP + out_spike_bytes
        else:
            return self.N_BYTES_FOR_TIMESTAMP + \
                        n_neurons * self.N_BYTES_PER_VALUE

    def get_buffered_sdram_per_timestep(self, variable, vertex_slice):
        """ Return the SDRAM used per timestep.

        In the case where sampling is used it returns the average\
        for recording and none recording based on the recording rate

        :param variable:
        :param vertex_slice:
        :return:
        """
        rate = self.__sampling_rates[variable]
        if rate == 0:
            return 0

        data_size = self.get_buffered_sdram_per_record(variable, vertex_slice)
        if rate == 1:
            return data_size
        else:
            return data_size // rate

    def get_sampling_overflow_sdram(self, vertex_slice):
        """ Get the extra SDRAM that should be reserved if using per_timestep

        This is the extra that must be reserved if per_timestep is an average\
        rather than fixed for every timestep.

        When sampling the average * time_steps may not be quite enough.\
        This returns the extra space in the worst case\
        where time_steps is a multiple of sampling rate + 1,\
        and recording is done in the first and last time_step

        :param vertex_slice:
        :return: Highest possible overflow needed
        """
        overflow = 0
        for variable, rate in iteritems(self.__sampling_rates):
            # If rate is 0 no recording so no overflow
            # If rate is 1 there is no overflow as average is exact
            if rate > 1:
                data_size = self.get_buffered_sdram_per_record(
                    variable,  vertex_slice)
                overflow += data_size // rate * (rate - 1)
        return overflow

    def get_buffered_sdram(self, variable, vertex_slice, n_machine_time_steps):
        """ Returns the SDRAM used for this may timesteps

        If required the total is rounded up so the space will always fit

        :param variable: The
        :param vertex_slice:
        :return:
        """
        rate = self.__sampling_rates[variable]
        if rate == 0:
            return 0
        data_size = self.get_buffered_sdram_per_record(variable, vertex_slice)
        records = n_machine_time_steps // rate
        if n_machine_time_steps % rate > 0:
            records = records + 1
        return data_size * records

    def get_sdram_usage_in_bytes(self, vertex_slice):
        n_words_for_n_neurons = (vertex_slice.n_atoms + 3) // 4
        n_bytes_for_n_neurons = n_words_for_n_neurons * 4
        return (8 + n_bytes_for_n_neurons) * len(self.__sampling_rates)

    def _get_fixed_sdram_usage(self, vertex_slice):
        total_neurons = vertex_slice.hi_atom - vertex_slice.lo_atom + 1
        fixed_sdram = 0
        # Recording rate for each neuron
        fixed_sdram += self.N_BYTES_PER_RATE
        # Number of recording neurons
        fixed_sdram += self.N_BYTES_PER_INDEX
        # index_parameters one per neuron
        # even if not recording as also act as a gate
        fixed_sdram += self.N_BYTES_PER_INDEX * total_neurons
        return fixed_sdram

    def get_variable_sdram_usage(self, vertex_slice):
        fixed_sdram = 0
        per_timestep_sdram = 0
        for variable in self.__sampling_rates:
            rate = self.__sampling_rates[variable]
            fixed_sdram += self._get_fixed_sdram_usage(vertex_slice)
            if rate > 0:
                fixed_sdram += self.SARK_BLOCK_SIZE
                per_record = self.get_buffered_sdram_per_record(
                    variable, vertex_slice)
                if rate == 1:
                    # Add size for one record as recording every timestep
                    per_timestep_sdram += per_record
                else:
                    # Get the average cost per timestep
                    average_per_timestep = per_record / rate
                    per_timestep_sdram += average_per_timestep
                    # Add the rest once to fixed for worst case
                    fixed_sdram += (per_record - average_per_timestep)
        return VariableSDRAM(fixed_sdram, per_timestep_sdram)

    def get_dtcm_usage_in_bytes(self, vertex_slice):
        # *_rate + n_neurons_recording_* + *_indexes
        usage = self.get_sdram_usage_in_bytes(vertex_slice)
        # *_count + *_increment
        usage += len(self.__sampling_rates) * self.N_BYTES_PER_POINTER * 2
        # out_spikes, *_values
        for variable in self.__sampling_rates:
            if variable == SPIKES:
                out_spike_words = int(math.ceil(vertex_slice.n_atoms / 32.0))
                out_spike_bytes = out_spike_words * self.N_BYTES_PER_WORD
                usage += self.N_BYTES_FOR_TIMESTAMP + out_spike_bytes
            else:
                usage += (self.N_BYTES_FOR_TIMESTAMP +
                          vertex_slice.n_atoms * self.N_BYTES_PER_VALUE)
        # *_size
        usage += len(self.__sampling_rates) * self.N_BYTES_PER_SIZE
        # n_recordings_outstanding
        usage += self.N_BYTES_PER_WORD * 4
        return usage

    def get_n_cpu_cycles(self, n_neurons):
        return n_neurons * self.N_CPU_CYCLES_PER_NEURON * \
                len(self.recording_variables)

    def get_data(self, vertex_slice):
        data = list()
        n_words_for_n_neurons = (vertex_slice.n_atoms + 3) // 4
        n_bytes_for_n_neurons = n_words_for_n_neurons * 4
        for variable in self.__sampling_rates:
            rate = self.__sampling_rates[variable]
            n_recording = self._count_recording_per_slice(
                variable, vertex_slice)
            data.append(numpy.array([rate, n_recording], dtype="uint32"))
            if rate == 0:
                data.append(numpy.zeros(n_words_for_n_neurons, dtype="uint32"))
            elif self.__indexes[variable] is None:
                data.append(numpy.arange(
                    n_bytes_for_n_neurons, dtype="uint8").view("uint32"))
            else:
                indexes = self.__indexes[variable]
                local_index = 0
                local_indexes = list()
                for index in xrange(n_bytes_for_n_neurons):
                    if index + vertex_slice.lo_atom in indexes:
                        local_indexes.append(local_index)
                        local_index += 1
                    else:
                        # write to one beyond recording range
                        local_indexes.append(n_recording)
                data.append(
                    numpy.array(local_indexes, dtype="uint8").view("uint32"))
        return numpy.concatenate(data)

    def get_global_parameters(self, vertex_slice):
        params = []
        for variable in self.__sampling_rates:
            params.append(NeuronParameter(
                self.__sampling_rates[variable], DataType.UINT32))
        for variable in self.__sampling_rates:
            n_recording = self._count_recording_per_slice(
                variable, vertex_slice)
            params.append(NeuronParameter(n_recording, DataType.UINT8))
        return params

    def get_index_parameters(self, vertex_slice):
        params = []
        for variable in self.__sampling_rates:
            if self.__sampling_rates[variable] <= 0:
                local_indexes = 0
            elif self.__indexes[variable] is None:
                local_indexes = IndexIsValue()
            else:
                local_indexes = []
                n_recording = sum(
                    vertex_slice.lo_atom <= index <= vertex_slice.hi_atom
                    for index in self.__indexes[variable])
                indexes = self.__indexes[variable]
                local_index = 0
                for index in xrange(
                        vertex_slice.lo_atom, vertex_slice.hi_atom+1):
                    if index in indexes:
                        local_indexes.append(local_index)
                        local_index += 1
                    else:
                        # write to one beyond recording range
                        local_indexes.append(n_recording)
            params.append(NeuronParameter(local_indexes, DataType.UINT8))
        return params

    @property
    def _indexes(self):  # for testing only
        return _ReadOnlyDict(self.__indexes)



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
from spinn_utilities.overrides import overrides
from spynnaker.pyNN.models.neuron.plasticity.stdp.common import (
    plasticity_helpers)
from .abstract_timing_dependence import AbstractTimingDependence
from spynnaker.pyNN.models.neuron.plasticity.stdp.synapse_structure import (
    SynapseStructureWeightOnly)

logger = logging.getLogger(__name__)

LOOKUP_TAU_PLUS_SIZE = 256
LOOKUP_TAU_PLUS_SHIFT = 0
LOOKUP_TAU_MINUS_SIZE = 256
LOOKUP_TAU_MINUS_SHIFT = 0


class TimingDependenceSpikePair(AbstractTimingDependence):
    __slots__ = [
        "__synapse_structure",
        "__tau_minus",
        "__tau_minus_last_entry",
        "__tau_plus",
        "__tau_plus_last_entry"]

    def __init__(self, tau_plus=20.0, tau_minus=20.0):
        self.__tau_plus = tau_plus
        self.__tau_minus = tau_minus

        self.__synapse_structure = SynapseStructureWeightOnly()

        # provenance data
        self.__tau_plus_last_entry = None
        self.__tau_minus_last_entry = None

    @property
    def tau_plus(self):
        return self.__tau_plus

    @property
    def tau_minus(self):
        return self.__tau_minus

    @overrides(AbstractTimingDependence.is_same_as)
    def is_same_as(self, timing_dependence):
        if not isinstance(timing_dependence, TimingDependenceSpikePair):
            return False
        return (self.__tau_plus == timing_dependence.tau_plus and
                self.__tau_minus == timing_dependence.tau_minus)

    @property
    def vertex_executable_suffix(self):
        return "pair"

    @property
    def pre_trace_n_bytes(self):

        # Pair rule requires no pre-synaptic trace when only the nearest
        # Neighbours are considered and, a single 16-bit R1 trace
        return 2

    @overrides(AbstractTimingDependence.get_parameters_sdram_usage_in_bytes)
    def get_parameters_sdram_usage_in_bytes(self):
        return 2 * (LOOKUP_TAU_PLUS_SIZE + LOOKUP_TAU_MINUS_SIZE)

    @property
    def n_weight_terms(self):
        return 1

    @overrides(AbstractTimingDependence.write_parameters)
    def write_parameters(self, spec, machine_time_step, weight_scales):
        # Check timestep is valid
        if machine_time_step != 1000:
            raise NotImplementedError(
                "STDP LUT generation currently only supports 1ms timesteps")

        # Write lookup tables
        self.__tau_plus_last_entry = plasticity_helpers.write_exp_lut(
            spec, self.__tau_plus, LOOKUP_TAU_PLUS_SIZE,
            LOOKUP_TAU_PLUS_SHIFT)
        self.__tau_minus_last_entry = plasticity_helpers.write_exp_lut(
            spec, self.__tau_minus, LOOKUP_TAU_MINUS_SIZE,
            LOOKUP_TAU_MINUS_SHIFT)

    @property
    def synaptic_structure(self):
        return self.__synapse_structure

    @overrides(AbstractTimingDependence.get_provenance_data)
    def get_provenance_data(self, pre_population_label, post_population_label):
        prov_data = list()
        prov_data.append(plasticity_helpers.get_lut_provenance(
            pre_population_label, post_population_label, "SpikePairRule",
            "tau_plus_last_entry", "tau_plus", self.__tau_plus_last_entry))
        prov_data.append(plasticity_helpers.get_lut_provenance(
            pre_population_label, post_population_label, "SpikePairRule",
            "tau_minus_last_entry", "tau_minus", self.__tau_minus_last_entry))
        return prov_data

    @overrides(AbstractTimingDependence.get_parameter_names)
    def get_parameter_names(self):
        return ['tau_plus', 'tau_minus']



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from spinn_utilities.overrides import overrides
from spynnaker.pyNN.models.neuron.plasticity.stdp.synapse_structure import (
    AbstractSynapseStructure)


class SynapseStructureWeightAccumulator(AbstractSynapseStructure):
    __slots__ = ()

    @overrides(AbstractSynapseStructure.get_n_half_words_per_connection)
    def get_n_half_words_per_connection(self):
        return 2

    @overrides(AbstractSynapseStructure.get_weight_half_word)
    def get_weight_half_word(self):
        return 0



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from spinn_utilities.overrides import overrides
from spynnaker.pyNN.external_devices_models import AbstractEthernetSensor
from .push_bot_translator import PushBotTranslator
from .push_bot_wifi_connection import get_pushbot_wifi_connection
from .push_bot_retina_connection import PushBotRetinaConnection
from spynnaker.pyNN.external_devices_models.push_bot import (
    AbstractPushBotRetinaDevice)


class PushBotEthernetRetinaDevice(
        AbstractPushBotRetinaDevice, AbstractEthernetSensor):
    def __init__(
            self, protocol, resolution, pushbot_ip_address, pushbot_port=56000,
            injector_port=None, local_host=None, local_port=None,
            retina_injector_label="PushBotRetinaInjector"):
        # pylint: disable=too-many-arguments
        super(PushBotEthernetRetinaDevice, self).__init__(protocol, resolution)
        pushbot_wifi_connection = get_pushbot_wifi_connection(
            pushbot_ip_address, pushbot_port)
        self.__translator = PushBotTranslator(
            protocol, pushbot_wifi_connection)
        self.__injector_port = injector_port
        self.__retina_injector_label = retina_injector_label

        self.__database_connection = PushBotRetinaConnection(
            self.__retina_injector_label, pushbot_wifi_connection, resolution,
            local_host, local_port)

    @overrides(AbstractEthernetSensor.get_n_neurons)
    def get_n_neurons(self):
        return self._resolution.value.n_neurons

    @overrides(AbstractEthernetSensor.get_injector_parameters)
    def get_injector_parameters(self):
        return {"port": self.__injector_port}

    @overrides(AbstractEthernetSensor.get_injector_label)
    def get_injector_label(self):
        return self.__retina_injector_label

    @overrides(AbstractEthernetSensor.get_translator)
    def get_translator(self):
        return self.__translator

    @overrides(AbstractEthernetSensor.get_database_connection)
    def get_database_connection(self):
        return self.__database_connection
